/*
 *  linux/arch/i386/kernel/signal.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  1997-11-28  Modified for POSIX.1b signals by Richard Henderson
 *  2000-06-20  Pentium III FXSR, SSE support by Gareth Hughes
 */

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/wait.h>
#include <linux/ptrace.h>
#include <linux/unistd.h>
#include <linux/stddef.h>
#include <linux/tty.h>
#include <linux/personality.h>
#include <asm/ucontext.h>
#include <asm/uaccess.h>
#include <asm/i387.h>
#include <asm/io.h>

#include <revirt/replay.h>
#include <asm/unistd.h>
#include <revirt/ztrace.h>


#define DEBUG_SIG 0

#include <linux/module.h>

typedef int (*signal_callback_sig)(struct pt_regs *regs, int signr);

static signal_callback_sig signal_callback = NULL;


void umlinux_reg_signal_callback(signal_callback_sig callback) {
  signal_callback = callback;
}

EXPORT_SYMBOL_NOVERS(umlinux_reg_signal_callback);


#define _BLOCKABLE (~(sigmask(SIGKILL) | sigmask(SIGSTOP)))

int FASTCALL(do_signal(struct pt_regs *regs, sigset_t *oldset));

int copy_siginfo_to_user(siginfo_t *to, siginfo_t *from)
{
#if 0
	if (!access_ok (VERIFY_WRITE, to, sizeof(siginfo_t)))
		return -EFAULT;
#endif
	if (from->si_code < 0)
		return __copy_to_user(to, from, sizeof(siginfo_t));
	else {
		int err=0;

		/* If you change siginfo_t structure, please be sure
		   this code is fixed accordingly.
		   It should never copy any pad contained in the structure
		   to avoid security leaks, but must copy the generic
		   3 ints plus the relevant union member.  */
                to->si_signo = from->si_signo;
                to->si_errno= from->si_errno ;
                to->si_code = from->si_code;
		/* First 32bits of unions are always present.  */
		to->si_pid=from->si_pid;
		switch (from->si_code >> 16) {
		case __SI_FAULT >> 16:
			break;
		case __SI_CHLD >> 16:
			to->si_utime=from->si_utime;
			to->si_stime=from->si_stime;
			to->si_status=from->si_status;
		default:
			to->si_uid=from->si_uid;
			break;
		/* case __SI_RT: This is not generated by the kernel as of now.  */
		}
		return err;
	}
}

/*
 * Atomically swap in the new signal mask, and wait for a signal.
 */
asmlinkage int
sys_sigsuspend(int history0, int history1, old_sigset_t mask)
{
	struct pt_regs * regs = (struct pt_regs *) &history0;
	sigset_t saveset;

	mask &= _BLOCKABLE;
	spin_lock_irq(&current->sigmask_lock);
	saveset = current->blocked;
	siginitset(&current->blocked, mask);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	regs->eax = -EINTR;
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		schedule();
		if (do_signal(regs, &saveset))
			return -EINTR;
	}
}

asmlinkage int
sys_rt_sigsuspend(sigset_t *unewset, size_t sigsetsize)
{
	struct pt_regs * regs = (struct pt_regs *) &unewset;
	sigset_t saveset, newset;

	/* XXX: Don't preclude handling different sized sigset_t's.  */
	if (sigsetsize != sizeof(sigset_t))
		return -EINVAL;

	if (copy_from_user(&newset, unewset, sizeof(newset)))
		return -EFAULT;
	sigdelsetmask(&newset, ~_BLOCKABLE);

	spin_lock_irq(&current->sigmask_lock);
	saveset = current->blocked;
	current->blocked = newset;
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	regs->eax = -EINTR;
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		schedule();
		if (do_signal(regs, &saveset))
			return -EINTR;
	}
}

asmlinkage int 
sys_sigaction(int sig, const struct old_sigaction *act,
	      struct old_sigaction *oact)
{
	struct k_sigaction new_ka, old_ka;
	int ret;

	if (act) {
		old_sigset_t mask;
		if (verify_area(VERIFY_READ, act, sizeof(*act)) ||
		    __get_user(new_ka.sa.sa_handler, &act->sa_handler) ||
		    __get_user(new_ka.sa.sa_restorer, &act->sa_restorer))
			return -EFAULT;
		__get_user(new_ka.sa.sa_flags, &act->sa_flags);
		__get_user(mask, &act->sa_mask);
		siginitset(&new_ka.sa.sa_mask, mask);
	}

	ret = do_sigaction(sig, act ? &new_ka : NULL, oact ? &old_ka : NULL);

	if (!ret && oact) {
		if (verify_area(VERIFY_WRITE, oact, sizeof(*oact)) ||
		    __put_user(old_ka.sa.sa_handler, &oact->sa_handler) ||
		    __put_user(old_ka.sa.sa_restorer, &oact->sa_restorer))
			return -EFAULT;
		__put_user(old_ka.sa.sa_flags, &oact->sa_flags);
		__put_user(old_ka.sa.sa_mask.sig[0], &oact->sa_mask);
	}

	return ret;
}

asmlinkage int
sys_sigaltstack(const stack_t *uss, stack_t *uoss)
{
	struct pt_regs *regs = (struct pt_regs *) &uss;
	return do_sigaltstack(uss, uoss, regs->esp);
}


/*
 * Do a signal return; undo the signal stack.
 */

struct sigframe
{
	char *pretcode;
	int sig;
	struct sigcontext sc;
	struct _fpstate fpstate;
	unsigned long extramask[_NSIG_WORDS-1];
	char retcode[8];
};

struct rt_sigframe
{
	char *pretcode;
	int sig;
	struct siginfo *pinfo;
	void *puc;
	struct siginfo info;
	struct ucontext uc;
	struct _fpstate fpstate;
	char retcode[8];
};

static int
restore_sigcontext(struct pt_regs *regs, struct sigcontext *sc, int *peax)
{
	unsigned int err = 0;

#define COPY(x)		err |= __get_user(regs->x, &sc->x)

#define COPY_SEG(seg)							\
	{ unsigned short tmp;						\
	  err |= __get_user(tmp, &sc->seg);				\
	  regs->x##seg = tmp; }

#define COPY_SEG_STRICT(seg)						\
	{ unsigned short tmp;						\
	  err |= __get_user(tmp, &sc->seg);				\
	  regs->x##seg = tmp|3; }

#define GET_SEG(seg)							\
	{ unsigned short tmp;						\
	  err |= __get_user(tmp, &sc->seg);				\
	  loadsegment(seg,tmp); }

	GET_SEG(gs);
	GET_SEG(fs);
	COPY_SEG(es);
	COPY_SEG(ds);
	COPY(edi);
	COPY(esi);
	COPY(ebp);
	COPY(esp);
	COPY(ebx);
	COPY(edx);
	COPY(ecx);
	COPY(eip);
	COPY_SEG_STRICT(cs);
	COPY_SEG_STRICT(ss);
	
	{
		unsigned int tmpflags;
		err |= __get_user(tmpflags, &sc->eflags);
		regs->eflags = (regs->eflags & ~0x40DD5) | (tmpflags & 0x40DD5);
		regs->orig_eax = -1;		/* disable syscall checks */
	}

	{
		struct _fpstate * buf;
		err |= __get_user(buf, &sc->fpstate);
		if (buf) {
			if (verify_area(VERIFY_READ, buf, sizeof(*buf)))
				goto badframe;
			err |= restore_i387(buf);
		}
	}

	err |= __get_user(*peax, &sc->eax);
	return err;

badframe:
	return 1;
}

asmlinkage int sys_sigreturn(unsigned long __unused)
{
	struct pt_regs *regs = (struct pt_regs *) &__unused;
	struct sigframe *frame = (struct sigframe *)(regs->esp - 8);
	sigset_t set;
	int eax;

	if (verify_area(VERIFY_READ, frame, sizeof(*frame)))
		goto badframe;
	if (__get_user(set.sig[0], &frame->sc.oldmask)
	    || (_NSIG_WORDS > 1
		&& __copy_from_user(&set.sig[1], &frame->extramask,
				    sizeof(frame->extramask))))
		goto badframe;

	sigdelsetmask(&set, ~_BLOCKABLE);
	spin_lock_irq(&current->sigmask_lock);
	current->blocked = set;
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	
	if (restore_sigcontext(regs, &frame->sc, &eax))
		goto badframe;
	return eax;

badframe:
	force_sig(SIGSEGV, current);
	return 0;
}	

asmlinkage int sys_rt_sigreturn(unsigned long __unused)
{
	struct pt_regs *regs = (struct pt_regs *) &__unused;
	struct rt_sigframe *frame = (struct rt_sigframe *)(regs->esp - 4);
	sigset_t set;
	stack_t st;
	int eax;

	if (verify_area(VERIFY_READ, frame, sizeof(*frame)))
		goto badframe;
	if (__copy_from_user(&set, &frame->uc.uc_sigmask, sizeof(set)))
		goto badframe;

	sigdelsetmask(&set, ~_BLOCKABLE);
	spin_lock_irq(&current->sigmask_lock);
	current->blocked = set;
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	
	if (restore_sigcontext(regs, &frame->uc.uc_mcontext, &eax))
		goto badframe;

	if (__copy_from_user(&st, &frame->uc.uc_stack, sizeof(st)))
		goto badframe;
	/* It is more difficult to avoid calling this function than to
	   call it and ignore errors.  */
	do_sigaltstack(&st, NULL, regs->esp);

	return eax;

badframe:
	force_sig(SIGSEGV, current);
	return 0;
}	

/*
 * Set up a signal frame.
 */

static int
setup_sigcontext(struct sigcontext *sc, struct _fpstate *fpstate,
		 struct pt_regs *regs, unsigned long mask)
{
	int tmp, err = 0;

	tmp = 0;
	__asm__("movl %%gs,%0" : "=r"(tmp): "0"(tmp));
	sc->gs=tmp;
	__asm__("movl %%fs,%0" : "=r"(tmp): "0"(tmp));
	sc->fs=tmp;
        sc->es = (unsigned int) regs->xes;
        sc->ds = (unsigned int) regs->xds;
        sc->edi = regs->edi;
        sc->esi = regs->esi;
        sc->ebp = regs->ebp;
        sc->esp = regs->esp;
        sc->ebx = regs->ebx;
        sc->edx = regs->edx;
        sc->ecx = regs->ecx;
        sc->eax = regs->eax;
        sc->trapno = current->thread.trap_no;
        sc->err = current->thread.error_code;
        sc->eip = regs->eip;
        sc->cs = (unsigned int) regs->xcs;
        /*
         * This affects the user's address space, so we need to make sure
         * it's consistent between runs.  A rogue process could detect whether
         * it's being replayed or not with this information.
         *
         * FIXME: check to make sure this doesn't break any programs.
         */
        sc->eflags = regs->eflags & (0x40DDF);
        sc->esp_at_signal = regs->esp;
        sc->ss = (unsigned int) regs->xss;


	tmp = save_i387(fpstate);
	if (tmp < 0)
	  err = 1;
	else
	  sc->fpstate = tmp ? fpstate : NULL;

	/* non-iBCS2 extensions.. */
	sc->oldmask = mask;
	sc->cr2 = current->thread.cr2;

	return err;
}

/*
 * Determine which stack to use..
 */
static inline void *
get_sigframe(struct k_sigaction *ka, struct pt_regs * regs, size_t frame_size)
{
	unsigned long esp;

	/* Default to using normal stack */
	esp = regs->esp;

	/* This is the X/Open sanctioned signal stack switching.  */
	if (ka->sa.sa_flags & SA_ONSTACK) {
		if (sas_ss_flags(esp) == 0)
			esp = current->sas_ss_sp + current->sas_ss_size;
	}

	/* This is the legacy signal stack switching. */
	else if ((regs->xss & 0xffff) != __USER_DS &&
		 !(ka->sa.sa_flags & SA_RESTORER) &&
		 ka->sa.sa_restorer) {
		esp = (unsigned long) ka->sa.sa_restorer;
	}

	return (void *)((esp - frame_size) & -8ul);
}

static void setup_frame(int sig, struct k_sigaction *ka,
			sigset_t *set, struct pt_regs * regs)
{
	struct sigframe *user_frame;
	struct sigframe tframe, * frame = &tframe;
	int err = 0;

	user_frame = get_sigframe(ka, regs, sizeof(*frame));

	/*
	 * Prevent random kernel stack things from affecting addrspace
	 */
	__clear_user(&tframe, sizeof(tframe));

	if (!access_ok(VERIFY_WRITE, user_frame, sizeof(*user_frame)))
		goto give_sigsegv;
	frame->sig= (current->exec_domain
		           && current->exec_domain->signal_invmap
		           && sig < 32
		           ? current->exec_domain->signal_invmap[sig]
		           : sig);

	err |= setup_sigcontext(&frame->sc, &frame->fpstate, regs, set->sig[0]);

	/* frame->sc.fpstate points to frame->fpstate if state is valid */
	if(frame->sc.fpstate) {
		frame->sc.fpstate = &user_frame->fpstate;
	}

	/* setup_sigcontext() returns an error if the FPU save fails. */
	if (err)
		goto give_sigsegv;

	/* __copy_to_user is often used as a bcopy to kernel space...*/
	if (_NSIG_WORDS > 1) {
		err |= __copy_to_user(frame->extramask, &set->sig[1],
				      sizeof(frame->extramask));
	}

	/* Set up to return from userspace.  If provided, use a stub
	   already in userspace.  */
	{
	  int i;
	  for(i=0; i<8; i++) 
	  	frame->retcode[i]='\0';
	}
	if (ka->sa.sa_flags & SA_RESTORER) {
		frame->pretcode=ka->sa.sa_restorer;
	} else {
		frame->pretcode=user_frame->retcode;
		/* This is popl %eax ; movl $,%eax ; int $0x80 */
		*(short *)(frame->retcode + 0) = 0xb858;
		*(int *)(frame->retcode+2) = __NR_sigreturn;
		*(short *)(frame->retcode+6) = 0x80cd;
	}

#ifdef CONFIG_REPLAY
        if((current->replay_flags & REPLAY_LOG)
                && ((1 << sig) & REPLAY_LOG_SIGNAL_MASK)) {
                log_sigframe((struct __kernel_sigframe * )frame);
        }
        if((current->replay_flags & REPLAY_REPLAY)
                && ((1 << sig) & REPLAY_LOG_SIGNAL_MASK)) {
           if(replay_sigframe((struct __kernel_sigframe *)user_frame,
                                (struct __kernel_sigframe *)frame) )
                goto give_sigsegv;
        } else  {
            if(__copy_to_user(user_frame, frame, sizeof(*frame)))
                goto give_sigsegv;
       }
#endif

	/* Set up registers for signal handler */
	regs->esp = (unsigned long) user_frame;
	regs->eip = (unsigned long) ka->sa.sa_handler;

	set_fs(USER_DS);
	regs->xds = __USER_DS;
	regs->xes = __USER_DS;
	regs->xss = __USER_DS;
	regs->xcs = __USER_CS;
	regs->eflags &= ~TF_MASK;

#if DEBUG_SIG
	printk("SIG deliver (%s:%d): sp=%p pc=%p ra=%p\n",
		current->comm, current->pid, user_frame, regs->eip, user_frame->pretcode);
#endif

	return;

give_sigsegv:
	if (sig == SIGSEGV)
		ka->sa.sa_handler = SIG_DFL;
	force_sig(SIGSEGV, current);
}

static void setup_rt_frame(int sig, struct k_sigaction *ka, siginfo_t *info,
			   sigset_t *set, struct pt_regs * regs)
{
	struct rt_sigframe *user_frame;
	struct rt_sigframe tframe, *frame = &tframe;
	int err = 0;

	user_frame = get_sigframe(ka, regs, sizeof(*frame));

	if (!access_ok(VERIFY_WRITE, user_frame, sizeof(*user_frame)))
		goto give_sigsegv;
	
	/*
	 * Prevent randmo kernel stack things from affecting userspace
	 */
	__clear_user(&tframe, sizeof(tframe));

	frame->sig= (current->exec_domain
		    	   && current->exec_domain->signal_invmap
		    	   && sig < 32
		    	   ? current->exec_domain->signal_invmap[sig]
			   : sig);
	frame->pinfo = &user_frame->info;
	frame->puc = &user_frame->uc;
	err |= copy_siginfo_to_user(&frame->info, info);
	if (err)
		goto give_sigsegv;

	/* Create the ucontext.  */
        frame->uc.uc_flags = 0;
        frame->uc.uc_link = 0;
        frame->uc.uc_stack.ss_sp = current->sas_ss_sp;
 
        frame->uc.uc_stack.ss_flags = sas_ss_flags(regs->esp);
 
        frame->uc.uc_stack.ss_size = current->sas_ss_size;
 
	err |= setup_sigcontext(&frame->uc.uc_mcontext, &frame->fpstate,
			        regs, set->sig[0]);

        /* frame->sc.fpstate points to frame->fpstate if state is valid */
        if(frame->uc.uc_mcontext.fpstate) {
                frame->uc.uc_mcontext.fpstate = &user_frame->fpstate;
        }
 
	err |= __copy_to_user(&frame->uc.uc_sigmask, set, sizeof(*set));
	if (err)
		goto give_sigsegv;

	/* Set up to return from userspace.  If provided, use a stub
	   already in userspace.  Clear it first. */
	{
	  int i;
	  for(i=0; i<8; i++)
	  	frame->retcode[i]='\0';
	}

	if (ka->sa.sa_flags & SA_RESTORER) {
		frame->pretcode=ka->sa.sa_restorer;
	} else {
		frame->pretcode=user_frame->retcode;
		/* This is movl $,%eax ; int $0x80 */
                *(char *)(frame->retcode+0) = 0xb8;
                *(int *)(frame->retcode + 1) = __NR_rt_sigreturn;
                *(short *)(frame->retcode+5) = 0x80cd;
	}

	if (err)
		goto give_sigsegv;

#ifdef CONFIG_REPLAY
        if(current->replay_flags & REPLAY_LOG
                && ((1 << sig) & REPLAY_LOG_SIGNAL_MASK)) {
                log_rt_sigframe((struct __kernel_rt_sigframe *) frame);
        }
 
        if((current->replay_flags & REPLAY_REPLAY)
                && ((1 << sig) & REPLAY_LOG_SIGNAL_MASK)) {
            if(replay_rt_sigframe((struct __kernel_rt_sigframe *)user_frame,
                                   (struct __kernel_rt_sigframe *)frame))
                goto give_sigsegv;
        } else  {
            if(__copy_to_user(user_frame, frame, sizeof(*frame)))
                goto give_sigsegv;
        }
#endif

	/* Set up registers for signal handler */
	regs->esp = (unsigned long) user_frame;
	regs->eip = (unsigned long) ka->sa.sa_handler;

	set_fs(USER_DS);
	regs->xds = __USER_DS;
	regs->xes = __USER_DS;
	regs->xss = __USER_DS;
	regs->xcs = __USER_CS;
	regs->eflags &= ~TF_MASK;

#if DEBUG_SIG
	printk("SIG deliver (%s:%d): sp=%p pc=%p ra=%p\n",
		current->comm, current->pid, user_frame, regs->eip, user_frame->pretcode);
#endif

	return;

give_sigsegv:
	if (sig == SIGSEGV)
		ka->sa.sa_handler = SIG_DFL;
	force_sig(SIGSEGV, current);
}


/*
 * OK, we're invoking a handler
 */	

static void
handle_signal(unsigned long sig, struct k_sigaction *ka,
	      siginfo_t *info, sigset_t *oldset, struct pt_regs * regs)
{
	/* Are we from a system call? */
	if (regs->orig_eax >= 0) {
		/* If so, check system call restarting.. */
		switch (regs->eax) {
			case -ERESTARTNOHAND:
				regs->eax = -EINTR;
				break;

			case -ERESTARTSYS:
				if (!(ka->sa.sa_flags & SA_RESTART)) {
					regs->eax = -EINTR;
					break;
				}
			/* fallthrough */
			case -ERESTARTNOINTR:
				regs->eax = regs->orig_eax;
				regs->eip -= 2;
		}
	}

        /*
         * We don't want the trap flag to be put into the sigcontext
         * and restored.  During logging, the trap flag shouldn't ever be set.
         * During replay, if we get a signal instead of a single-step, we
         * want to yield after the signal is delivered, and the trap flag
         * to be cleared.
         */
        if(current->replay_flags & REPLAY_REPLAY) {
                regs->eflags &= ~TF_MASK;
        }

	/* Set up the stack frame */
	if (ka->sa.sa_flags & SA_SIGINFO)
		setup_rt_frame(sig, ka, info, oldset, regs);
	else
		setup_frame(sig, ka, oldset, regs);

	if (ka->sa.sa_flags & SA_ONESHOT)
		ka->sa.sa_handler = SIG_DFL;

	if (!(ka->sa.sa_flags & SA_NODEFER)) {
		spin_lock_irq(&current->sigmask_lock);
		sigorsets(&current->blocked,&current->blocked,&ka->sa.sa_mask);
		sigaddset(&current->blocked,sig);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);
	}
}

/*
 * Note that 'init' is a special process: it doesn't get signals it doesn't
 * want to handle. Thus you cannot kill init even with a SIGKILL even by
 * mistake.
 */
int do_signal(struct pt_regs *regs, sigset_t *oldset)
{
	siginfo_t info;
	struct k_sigaction *ka;
	int first=1;
	int rflags, ret;

	/*
	 * We want the common case to go fast, which
	 * is why we may in certain cases get here from
	 * kernel mode. Just return without doing anything
	 * if so.
	 */
	if ((regs->xcs & 3) != 3) {
	  return 1;
	}

	if (!oldset)
		oldset = &current->blocked;

        /*
         * do_signal() may be called from sys_sigsuspend; so we need to
         * save the state of replay_flags and clear XUSER and DATASQUASH
         * so it doesn't confound us.
         */
        rflags = current->replay_flags & (REPLAY_XUSER | REPLAY_DATASQUASH);
        current->replay_flags &= ~(REPLAY_XUSER | REPLAY_DATASQUASH);
 
	for (;;) {
		unsigned long signr;
#ifdef CONFIG_REPLAY
                if(current->replay_flags & REPLAY_LOG && !first) {
                        printk("do_signal: first not set! [pause]\n");
                        ztrace_debug_pause();
                }
#endif

		spin_lock_irq(&current->sigmask_lock);
		signr = dequeue_signal(&current->blocked, &info);
		spin_unlock_irq(&current->sigmask_lock);

		if (!signr)
			break;

		if ((current->ptrace & PT_PTRACED) && signr != SIGKILL) {
#ifdef CONFIG_REPLAY
			if(current->umlinux.umlTraced) {
			    if(signal_callback!=NULL) {
				current->exit_code=signal_callback(regs, signr);
			    }
			} else {
#else
			{
#endif
			    /* Let the debugger run.  */
			    current->exit_code = signr;
			    current->state = TASK_STOPPED;
			    notify_parent(current, SIGCHLD);
			    schedule();
			}

			/* We're back.  Did the debugger cancel the sig?  */
			if (!(signr = current->exit_code))
				continue;
			current->exit_code = 0;

			/* The debugger continued.  Ignore SIGSTOP.  */
			if (signr == SIGSTOP)
				continue;

			/* Update the siginfo structure.  Is this good?  */
			if (signr != info.si_signo) {
				info.si_signo = signr;
				info.si_errno = 0;
				info.si_code = SI_USER;
				info.si_pid = current->p_pptr->pid;
				info.si_uid = current->p_pptr->uid;
			}

			/* If the (new) signal is now blocked, requeue it.  */
			if (sigismember(&current->blocked, signr)) {
				send_sig_info(signr, &info, current);
				continue;
			}
		}

#ifdef REPLAY_DEBUG
                if(current->replay_flags & REPLAY_REPLAY) {
                   unsigned long code[4];
                   unsigned long * eip = (unsigned long *)regs->eip;
                   int i;
                   printk("do_signal: delivering signal %d to pid %d @ 0x%x\n",
                               signr, current->pid, regs->eip);
                   for(i=0; i<4; i++) {
                        get_user(code[i], eip);
                        eip += 4;
                   }
                   printk("do_signal: code %x %x %x %x\n",
                                code[0],
                                code[1],
                                code[2],
                                code[3]);
                }
#endif


		ka = &current->sig->action[signr-1];
		if (ka->sa.sa_handler == SIG_IGN) {
			if (signr != SIGCHLD)
				continue;
			/* Check for SIGCHLD: it's special.  */
			while (sys_wait4(-1, NULL, WNOHANG, NULL) > 0)
				/* nothing */;
			continue;
		}

#ifdef CONFIG_REPLAY
                /*
                 * real "Handling" begins here, so here's where we log...
                 */
                if(current->replay_flags & REPLAY_LOG) {
                    if((1 << signr) & REPLAY_LOG_SIGNAL_MASK) {
                        save_perfctr(current);
                        log_signal(&info);
                    }
                }
# ifdef REPLAY_DEBUG
                else if(current->replay_flags & REPLAY_REPLAY
                        && ((1<<signr) & REPLAY_LOG_SIGNAL_MASK)) {
                    printk("do_signal: pid %d handling logged signal %d\n",
                                current->pid, signr);
                }
# endif
#endif	
		if (ka->sa.sa_handler == SIG_DFL) {
			int exit_code = signr;

			/* Init gets no signals it doesn't want.  */
			if (current->pid == 1)
				continue;

			switch (signr) {
			case SIGCONT: case SIGCHLD: case SIGWINCH:
				continue;

			case SIGTSTP: case SIGTTIN: case SIGTTOU:
				if (is_orphaned_pgrp(current->pgrp))
					continue;
				/* FALLTHRU */

			case SIGSTOP: {
				struct signal_struct *sig;
				current->state = TASK_STOPPED;
				current->exit_code = signr;
				sig = current->p_pptr->sig;
				if (sig && !(sig->action[SIGCHLD-1].sa.sa_flags & SA_NOCLDSTOP))
					notify_parent(current, SIGCHLD);
				schedule();
				continue;
			}

			case SIGQUIT: case SIGILL: case SIGTRAP:
			case SIGABRT: case SIGFPE: case SIGSEGV:
			case SIGBUS: case SIGSYS: case SIGXCPU: case SIGXFSZ:
				if (do_coredump(signr, regs))
					exit_code |= 0x80;
				/* FALLTHRU */

			default:
				sigaddset(&current->pending.signal, signr);
				recalc_sigpending(current);
				current->flags |= PF_SIGNALED;
                                /*
                                 * FIXME - we should probably do this
                                 * for other signals too -- SIGSTOP?
                                 */
#ifdef CONFIG_REPLAY
                                if(current->replay_flags & REPLAY_REPLAY)
                                    ztrace_tsk_gets_signal();
#endif
				do_exit(exit_code);
				/* NOTREACHED */
			}
		}

		/* Reenable any watchpoints before delivering the
		 * signal to user space. The processor register will
		 * have been cleared if the watchpoint triggered
		 * inside the kernel.
		 */
		__asm__("movl %0,%%db7"	: : "r" (current->thread.debugreg[7]));

		/* Whee!  Actually deliver the signal.  */
		handle_signal(signr, ka, &info, oldset, regs);
#ifdef CONFIG_REPLAY
		if(current->replay_flags & REPLAY_REPLAY) {
			ztrace_tsk_gets_signal();
		}
#endif
		ret=1;
		goto out;
	}

	/* Did we come from a system call? */
	if (regs->orig_eax >= 0) {
		/* Restart the system call - no handlers present */
		if (regs->eax == -ERESTARTNOHAND ||
		    regs->eax == -ERESTARTSYS ||
		    regs->eax == -ERESTARTNOINTR) {
			regs->eax = regs->orig_eax;
			regs->eip -= 2;
		}
	}
	ret=0;
out:

	return ret;
}
