/*
 * lexecve.c - Log executable using the ReVirt system
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/perfctr.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <revirt/rlog.h>


#define MAX_ARG_SPACE 2*4096

static int log_exec_params(struct task_struct* task, char *fname,char** argv,
								char **env);
extern inline void clear_branchdata(struct task_struct *tsk);

/*
 * sys_lexecve() sets the logging flag of the process and
 * executes a new program. 
 * Synopsis:
 * lexecve( char *filename, char **argp, char **envp, 
 *                                               char *syscall_log_filename )
 */
asmlinkage int sys_lexecve(struct pt_regs regs)
{
	char * filename;
	int error;
        struct rlog_schedule_entry le;

        /* perhaps we should do some checking, for the newer libc. */
	printk("sys_lexecve is called\n");
	lock_kernel();

	current->replay_flags= REPLAY_LOG;
	current->replay_nrex= 0;

#ifdef REPLAY_UNIQID
	rlog_resetid(0);
#endif

        le= (struct rlog_schedule_entry) { 
                .type= ENTRY_TOP,
                .pid= current->pid,
        	.data= 0
              };

	demon_write_log((char *)&le, le.type, sizeof(le));

	perfctr_enable_counters();
	perfctr_clear_counters();

	clear_branchdata(current);

	error= log_exec_params(current,(char*) regs.ebx,(char**) regs.ecx,
	                 (char**) regs.edx);
	if (error)
		goto out;

	/* rest is from the original execve code */
	filename = getname((char *) regs.ebx);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out;
	error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, 
							&regs);

out_fname:
	putname(filename);
	
	unlock_kernel();
	return error;
out:
	current->replay_flags= 0;
	unlock_kernel();
	return error;
}

/*
 * We need to know the size of the buffer so that we can ensure we have no
 * buffer overruns -- i.e., we need strnlen_user to make sense.
 */
static int copy_argv(char *as, char **argv, int size) {
	char * arg;
	int argsize, i;

	for(i=0; ; i++) {
		arg = 0;
		copy_from_user(&arg, argv+i, 4);
		if(arg == 0) {
			return size;
		}
		argsize=strnlen_user(arg, MAX_ARG_SPACE-size-1);
		if(arg[argsize-1]!=0) 
			return -E2BIG;
		copy_from_user(as, arg, argsize);
		size += argsize;
		as += argsize;
	}
}


static int log_exec_params(struct task_struct *task, char *fname,
				char** argv, char **env) {
	int size, fsize;
	char *buf, *a, *e, *f;
	int error;
	struct rlog_entry_scall_lexecve * se;

	/* FIXME: do PTR_ERR checks here */
	fsize=strnlen_user(fname, MAX_ARG_SPACE-1);
	error= -E2BIG;
	if (fname[fsize-1]!=0)
		goto out;

	error= -ENOMEM;
	buf= kmalloc(MAX_ARG_SPACE, GFP_KERNEL);
	if (!buf)
		goto out;
	
	se = (struct rlog_entry_scall_lexecve *)buf;
	se->type= ENTRY_EXECPARM;
	se->pid = current->pid;
	size= sizeof(struct rlog_entry_scall_lexecve);

	f= buf + size;
	copy_from_user(f, fname, fsize);
	size += fsize;

	a = buf + size;
	se->argv_index = size;
	size = copy_argv(a, argv, size);
	if (size<0) {
		error= size;
		goto out_free_arg_space;
	}


	e = buf+size;
	se->env_index = size;
	size = copy_argv(e, env, size);
	if (size<0) {
		error= size;
		goto out_free_arg_space;
	}

	demon_write_log(buf, ENTRY_EXECPARM, size);
	error=0;

out_free_arg_space:
	kfree(buf);

out:
	return error;
}
