/*
 * perfctr.c - Provide abstract access to arch-dependent hardware counters
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/types.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/perfctr.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <revirt/rlog.h>

#if defined(CONFIG_M686) || defined(CONFIG_MPENTIUMIII) 

/************************** P6 CODE ************************************/


#define P6_EV_INST_RETIRED		0xc0
#define P6_EV_BR_TAKEN_RETIRED		0xc9
#define P6_EV_HW_INT_RX			0xc8

#define P6_USRMODE	 		(1<<16)
#define P6_OSMODE			(1<<17)
#define P6_INT				(1<<20)
#define P6_EN				(1<<22)

#define P6_EVTSEL			0x186
#define P6_EVTSEL0			0x186
#define P6_EVTSEL1			0x187

#define P6_CTR				0xc1
#define P6_CTR0				0xc1
#define P6_CTR1				0xc2

static void perfctr_set_evtsel(int i, struct intstruct x) {
	wrmsr(P6_EVTSEL + i, x.lo, x.hi);
}

static void perfctr_get_evtsel(int i, struct intstruct * x) {
	rdmsr(P6_EVTSEL + i, x->lo, x->hi);
}

void perfctr_set_ctr(int i, s64 x) {
	u32 hi, lo;
	lo = (u32) x;
	hi = (u32) ((x-lo) >> 32) ;
	wrmsr(P6_CTR + i, lo, hi);
}

void perfctr_get_ctr(int i, s64 * x) {
        u32 hi,lo;
        rdmsr(P6_CTR+i, lo, hi);
        /*
         * P6 perfctrs are only 48 bits, so we need to
         * extend hi from 4 to 32 bits if negative
         */
        if(hi & 0x8) {
            hi |= 0xFFFFFFF0L;
        }
        *x = ((s64)(hi) << 32) + lo;
}

void save_perfctr_regs(struct task_struct * prev) {
        int i;
        for(i=0; i<2; i++)  {
            perfctr_get_ctr(i, &prev->perfctr[i].ctr);
        }
	prev->replay_int=replay_interrupts;
	prev->replay_ex=replay_exceptions;
}

void restore_perfctr_regs(struct task_struct * next) {
        int i;
        for(i=0; i<2; i++)  {
            perfctr_set_ctr(i, next->perfctr[i].ctr);
        }
	replay_interrupts=next->replay_int;
	replay_exceptions=next->replay_ex;
}

void perfctr_enable_counters(void) {
	wrmsr(P6_EVTSEL0, (P6_EV_BR_TAKEN_RETIRED | P6_USRMODE | P6_EN), 0);
	wrmsr(P6_EVTSEL1, (P6_EV_HW_INT_RX | P6_USRMODE | P6_EN), 0);
}

void perfctr_clear_counters(void) {
	wrmsr(P6_CTR0,0,0);
	wrmsr(P6_CTR1,0,0);
	replay_interrupts=0;
	replay_exceptions=0;
}

void perfctr_enable_counters_tsk(struct task_struct * tsk) {
	tsk->perfctr[0].evtsel.lo = 
			(P6_EV_BR_TAKEN_RETIRED | P6_USRMODE | P6_EN);
	tsk->perfctr[0].evtsel.hi = 0;
	tsk->perfctr[1].evtsel.lo = 
		   (P6_EV_HW_INT_RX | P6_USRMODE | P6_EN);
	tsk->perfctr[1].evtsel.hi = 0;
}

void perfcntr_enable_overflow_tsk(struct task_struct * tsk) {
	int l,h;

        l=apic_read(APIC_LVTPC);
        l &= ~(APIC_LVT_MASKED);
        apic_write_around(APIC_LVTPC, l);

        rdmsr(P6_EVTSEL0, l, h);
        l |= P6_INT;
        wrmsr(P6_EVTSEL0, l, h);
}

void perfcntr_disable_overflow_tsk(struct task_struct * tsk) {
        int l, h;
        rdmsr(P6_EVTSEL0, l, h);
        l &= ~P6_INT;
        wrmsr(P6_EVTSEL0, l, h);
}


#elif defined(CONFIG_MK6) || defined(CONFIG_MK7)

/************************ AMD ATHLON CODE *******************************/


#define ATH_EV_INST_RETIRED		0xc0
#define ATH_EV_BR_TAKEN_RETIRED		0xc4
#define ATH_EV_HW_INT_RX		0xcf

#define ATH_USRMODE	 		(1<<16)
#define ATH_OSMODE			(1<<17)
#define ATH_INT				(1<<20)
#define ATH_EN				(1<<22)

#define ATH_EVTSEL			0xC0010000
#define ATH_EVSEL0			0xC0010000
#define ATH_EVSEL1			0xC0010001

#define ATH_CTR				0xC0010004
#define ATH_CTR0			0xC0010004
#define ATH_CTR1			0xC0010005

/*
 * This is generalized code and should be in arch/
 */
static void perfctr_set_evtsel(int i, struct intstruct x) {
	wrmsr(ATH_EVTSEL + i, x.lo, x.hi);
}

static void perfctr_get_evtsel(int i, struct intstruct * x) {
	rdmsr(ATH_EVTSEL+i, x->lo, x->hi);
}

static void perfctr_set_ctr(int i, s64 x) {
	u32 hi,lo;
	lo = (u32) x;
	hi = (u32) ((x-lo) >> 32);
	wrmsr(ATH_CTR + i, lo, hi);
}

static void perfctr_get_ctr(int i, s64 * x) {
	u32 hi,lo;
	rdmsr(ATH_CTR+i, lo, hi);
	/*
	 * Athalon perfctrs are only 48 bits, so we need to 
	 * extend hi from 16 to 32 bits if negative
	 */
	if(hi & 0x8000) {
	    hi |= 0xFFFF0000L;
	}
	*x = ((s64)(hi) << 32) + lo;
}


// interface
void save_perfctr_regs(struct task_struct * prev) {
	int i;
	for(i=0; i<2; i++)  {
	    perfctr_get_ctr(i, &prev->perfctr[i].ctr);
	}
	prev->replay_int=replay_interrupts;
	prev->replay_ex=replay_exceptions1;
}
// interface
void restore_perfctr_regs(struct task_struct * next) {
	int i;
	for(i=0; i<2; i++)  {
	    perfctr_set_ctr(i, next->perfctr[i].ctr);
	}
	replay_interrupts=next->replay_int;
	replay_exceptions1=next->replay_ex;
	replay_exceptions2=0;

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	// This is sort of a hack; we don't need precise counts, we just
	// want to know sort of what's happening.
	for(i=0; i<24; i++)
	    replay_ecount[i]=0;
#endif
}


/*
 * This is revirt-specific code and should be in revirt/
 */
void perfctr_enable_counters(void) {
	wrmsr(ATH_EVSEL0,(ATH_EV_BR_TAKEN_RETIRED| ATH_USRMODE | ATH_EN), 0 );
	wrmsr(ATH_EVSEL1,(ATH_EV_HW_INT_RX | ATH_USRMODE | ATH_EN), 0 );
}

void perfctr_clear_counters(void) {
	wrmsr(ATH_CTR0,0,0);
	wrmsr(ATH_CTR1,0,0);
	replay_interrupts=0;
	replay_exceptions1 = replay_exceptions2=0;
}

// interface
void perfctr_enable_counters_tsk(struct task_struct * tsk) {
	tsk->perfctr[0].evtsel.lo = 
			     (ATH_EV_BR_TAKEN_RETIRED| ATH_USRMODE | ATH_EN);
	tsk->perfctr[0].evtsel.hi=0;
	tsk->perfctr[1].evtsel.lo=(ATH_EV_HW_INT_RX | ATH_USRMODE | ATH_EN);
	tsk->perfctr[1].evtsel.hi=0;
}

void perfcntr_enable_overflow_tsk(struct task_struct * tsk) {
	int l, h;

        l=apic_read(APIC_LVTPC);
	l &= ~(APIC_LVT_MASKED);
	apic_write_around(APIC_LVTPC, l);

	rdmsr(ATH_EVSEL0, l, h);
	l |= ATH_INT;
	wrmsr(ATH_EVSEL0, l, h);
}

void perfcntr_disable_overflow_tsk(struct task_struct * tsk) {
	int l, h;
	rdmsr(ATH_EVSEL0, l, h);
	l &= ~ATH_INT;
	wrmsr(ATH_EVSEL0, l, h);
}

#elif defined(CONFIG_MPENTIUM4)
/**************************  P4 CODE ********************************/


#define MSR_IQ_COUNTER0			0x30c
#define MSR_IQ_CCCR0			0x36c
#define MSR_CRU_ESCR2			0x3cc

#define ESCR_USR			(1<<2)
#define ESCR_EVSEL(x)			(((x)&((1<<6)-1))<<25)
#define ESCR_EVMASK(x)			(((x)&((1<<16)-1))<<9)

#define CCCR_EN				(1<<12)
/* FIXME -- why are we setting these reserved bits? */
#define CCCR_RESERVED			(3<<16)
#define CCCR_OVF_PMI_T0			(1<<26)
#define CCCR_OVF			(1<<31)
#define CCCR_ESCR_SEL(x)		(((x)&7)<<13)

/* 
 * FIXME  -- for now, since high bits are always zero...
 */
void perfctr_set_evtsel(struct intstruct x) {
	wrmsr(MSR_CRU_ESCR2, x.lo, 0);
	wrmsr(MSR_IQ_CCCR0, x.hi, 0);
}

void perfctr_get_evtsel(struct intstruct * x) {
	int nul;
	rdmsr(MSR_CRU_ESCR2, x->lo, nul);
	rdmsr(MSR_IQ_CCCR0, x->hi, nul);
}

void perfctr_set_ctr(s64 x) { 
	u32 hi, lo;
	lo = (u32) x;
	hi = (u32) ((x-lo) >> 32);
        wrmsr(MSR_IQ_COUNTER0, lo, hi);
}

void perfctr_get_ctr(s64 * x) {
	u32 hi, lo;
        rdmsr(MSR_IQ_COUNTER0, lo, hi);
	/*
	 * P4 perfctrs are only 40 bits, so we need to 
	 * extend hi from 8 to 32 bits if negative
	 */
	if(hi & 0x80) {
	    hi |= 0xFFFFFF00L;
	}
	*x = ((s64)(hi) << 32) + lo;
}


// interface
void save_perfctr_regs(struct task_struct * prev) {
        perfctr_get_ctr(&prev->perfctr[0].ctr);
//	perfctr_get_evtsel(&prev->perfctr[0].evtsel);
	prev->replay_int=replay_interrupts;
	prev->replay_ex=replay_exceptions1;
}
// interface
void restore_perfctr_regs(struct task_struct * next) {
//	perfctr_set_evtsel(next->perfctr[0].evtsel);
        perfctr_set_ctr(next->perfctr[0].ctr);
	replay_interrupts=next->replay_int;
	replay_exceptions1=next->replay_ex;
}

/*
 * Sukru had it set up to count *all* branches -- taken and not 
 * (evtmask == 15).  It shouldn't be broken, so don't fix it...
 */
void perfctr_enable_counters(void) {
	wrmsr(MSR_CRU_ESCR2, ESCR_USR| ESCR_EVSEL(6) | ESCR_EVMASK(15), 0);
	wrmsr(MSR_IQ_CCCR0, CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(5), 0 );
}

void perfctr_clear_counters(void) {
	wrmsr(MSR_IQ_COUNTER0, 0,0);
	replay_interrupts=0;
	replay_exceptions1=0;
}

// interface
void perfctr_enable_counters_tsk(struct task_struct * tsk) {
	tsk->perfctr[0].evtsel.lo = ESCR_USR| ESCR_EVSEL(6) | ESCR_EVMASK(15);
	tsk->perfctr[0].evtsel.hi = CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(5);
	tsk->perfctr[1].evtsel.lo=0;
	tsk->perfctr[1].evtsel.hi=0;
}

// interface
void perfcntr_enable_overflow_tsk(struct task_struct * tsk) {
	int l, h;

	/*
	 * FIXME - Should OVF clearing and LVTPC masking be dealt with in the 
	 * interrupt handler?
	 */
	l=apic_read(APIC_LVTPC);
	l &= ~(APIC_LVT_MASKED);
	apic_write_around(APIC_LVTPC, l);

	rdmsr(MSR_IQ_CCCR0, l, h);
	l &= ~(CCCR_OVF);
	l |= CCCR_OVF_PMI_T0;
	wrmsr(MSR_IQ_CCCR0, l, h);
}

// interface
void perfcntr_disable_overflow_tsk(struct task_struct * tsk) {
	int l, h;
	rdmsr(MSR_IQ_CCCR0, l, h);
	l &= ~(CCCR_OVF_PMI_T0);
	wrmsr(MSR_IQ_CCCR0, l, h);
}



#else
#error "don't know perfctrs for this architecture"
#endif
