/*
 * replay.c - ReVirt replay functionality
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/linkage.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/tqueue.h>
#include <linux/slab.h>
#include <asm/perfctr.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/debugreg.h>
#include <asm/user.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>
#include <asm/unistd.h>

#include <asm/fixmap.h>
#include <asm/apic.h>
#include <asm/mtrr.h>
#include <asm/siginfo.h>

//#define REPLAY_DEBUG

int replay_sigframe(struct __kernel_sigframe *user_frame, 
				struct __kernel_sigframe *frame){
	struct rlog_entry_sigframe * e;
	struct rlog_header rh;
	int pause, i;

	demon_read_log(&rh);
	if(rh.type != ENTRY_SIGFRAME) {
		printk("replay_sigframe: expecting type %x, got %x (%u)\n",
				ENTRY_SIGFRAME, rh.type, rh.id);
	}
	e=(struct rlog_entry_sigframe *)replay_scall_buf;

	pause=0;
#define LDIFF(x,y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("replay_sigframe: %s differs: %lx %lx\n", \
	  	y, frame->x, e->frame.x);	  	 \
  	  pause = 1;					 \
	}
#define DIFF(x,y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("replay_sigframe: %s differs: %x %x\n", \
	  	y, frame->x, e->frame.x);	  	 \
  	  pause = 1;					 \
	}
	DIFF(pretcode, "pretcode");
	DIFF(sig, "sig");
	LDIFF(sc.eax, "eax");
	LDIFF(sc.ebx, "ebx");
	LDIFF(sc.ecx, "ecx");
	LDIFF(sc.edx, "edx");

	LDIFF(sc.edi, "edi");
	LDIFF(sc.esi, "esi");
	LDIFF(sc.ebp, "ebp");
	LDIFF(sc.esp, "esp");
	LDIFF(sc.eip, "eip");
	DIFF(sc.cs, "cs");
	DIFF(sc.ss, "ss");
	DIFF(sc.gs, "gs");
	DIFF(sc.fs, "fs");
	DIFF(sc.es, "es");
	DIFF(sc.ds, "ds");
#undef DIFF
#undef LDIFF
	/*
	 * only the bits masked below are restored by restore_sigcontext()
	 * in arch/i386/kernel/signal.c.
	 * I suppose we'd really prefer to check the actual flags
	 * restored during log and replay, but this will have to do for now...
	 */
	if((e->frame.sc.eflags & 0x40DD5) 
		!= (frame->sc.eflags & 0x40DD5)) {
	    printk("replay_sigframe: restored eflags differ: 0x%lx 0x%lx\n",
			frame->sc.eflags & 0x40DD5,
	    		e->frame.sc.eflags & 0x40DD5);
	    
	    pause=1;
	}

	for(i=0; i<8; i++) {
	 if(e->frame.retcode[i] != frame->retcode[i]) {
	   printk("replay_sigframe: retcode[%d] differs: %x %x\n",
	   		i, frame->retcode[i], e->frame.retcode[i]);
	   pause=1;
	 }
	}


     	if(pause) {
     	  printk("   pausing...\n");
	  ztrace_debug_pause();
     	}

	return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
}

int replay_rt_sigframe(struct __kernel_rt_sigframe *user_frame, 
			struct __kernel_rt_sigframe *frame){
	struct rlog_entry_rt_sigframe * e;
	struct rlog_header rh;
	int pause, i;

	demon_read_log(&rh);
	if(rh.type != ENTRY_RT_SIGFRAME) {
		printk("replay_rt_sigframe: expecting logtype %x, got %x(%d)\n",
				ENTRY_RT_SIGFRAME, rh.type, rh.id);
	}
	e=(struct rlog_entry_rt_sigframe *)replay_scall_buf;

#ifdef REPLAY_DEBUG
	printk("replay_rt_sigframe: handling signal %x\n", frame->info.si_signo);
#endif

	pause=0;
#ifdef REPLAY_WHOLE_SIGFRAME
#define DIFF(x, y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("replay_rt_sigframe: %s differs: %lx %lx\n", \
	  	y, frame->x, e->frame.x);	  	 \
  	  pause = 1;					 \
	}
	DIFF(pretcode, "pretcode");
	DIFF(pinfo, "pinfo");
	DIFF(puc, "puc");

	DIFF(info.si_signo, "info.si_signo");
	DIFF(info.si_errno, "info.si_errno");
	DIFF(info.si_code, "info.si_code");
	DIFF(info.si_errno, "info.si_errno");

	// This is probably OK, as long as we clear the structure first
	// i.e., I suspect the differences are just random stack differences,
	// not true changes.  Nevertheless, we need it to be the same both
	// times.
	for(i=0;i < SI_PAD_SIZE; i++) {
	 if(e->frame.info._sifields._pad[i] != frame->info._sifields._pad[i]) {
	   printk("replay_rt_sigframe: frame.info._sifields._pad[%d] differs: %x %x\n",
	   		i, 
			frame->info._sifields._pad[i],
	 		e->frame.info._sifields._pad[i]);
	  pause=1;
	 }
	}

	DIFF(uc.uc_flags, "uc.uc_flags");
	DIFF(uc.uc_link, "uc.uc_link");
	DIFF(uc.uc_stack.ss_sp, "uc.uc_stack.ss_sp");
	DIFF(uc.uc_stack.ss_flags, "uc.uc_stack.ss_flags");
	DIFF(uc.uc_stack.ss_size, "uc.uc_stack.ss_size");
	
	DIFF(uc.uc_mcontext.gs, "gs");
	DIFF(uc.uc_mcontext.fs, "fs");
	DIFF(uc.uc_mcontext.es, "es");
	DIFF(uc.uc_mcontext.ds, "ds");
	DIFF(uc.uc_mcontext.__fsh, "__fsh");
	DIFF(uc.uc_mcontext.__esh, "__esh");
	DIFF(uc.uc_mcontext.__dsh, "__dsh");
	DIFF(uc.uc_mcontext.__csh, "__csh");
	DIFF(uc.uc_mcontext.edi, "edi");
	DIFF(uc.uc_mcontext.esi, "esi");
	DIFF(uc.uc_mcontext.ebp, "ebp");
	DIFF(uc.uc_mcontext.esp, "esp");
	DIFF(uc.uc_mcontext.ebx, "ebx");
	DIFF(uc.uc_mcontext.ecx, "ecx");
	DIFF(uc.uc_mcontext.edx, "edx");
	DIFF(uc.uc_mcontext.eax, "eax");
	// We know that trapno changes
	DIFF(uc.uc_mcontext.err, "err");
	DIFF(uc.uc_mcontext.eip, "eip");
	DIFF(uc.uc_mcontext.cs, "cs");
	// eflags checked below
	DIFF(uc.uc_mcontext.esp_at_signal, "esp_at_signal");
	DIFF(uc.uc_mcontext.ss, "ss");
	DIFF(uc.uc_mcontext.fpstate, "* fpstate");
	DIFF(uc.uc_mcontext.oldmask, "oldmask");
	DIFF(uc.uc_mcontext.cr2, "cr2");
	
	/*
	 * only the bits masked below are restored by restore_sigcontext.
	 * I suppose we'd really prefer to check the actual flags
	 * restored during log and replay, but this will have to do for now...
	 */
	if((e->frame.uc.uc_mcontext.eflags & 0x40DD5) 
		!= (frame->uc.uc_mcontext.eflags & 0x40DD5)) {
	    printk("replay_rt_sigframe: restored eflags differ: 0x%lx 0x%lx\n",
			frame->uc.uc_mcontext.eflags & 0x40DD5,
	    		e->frame.uc.uc_mcontext.eflags & 0x40DD5);
	    pause=1;
	}
	for(i=0; i<8; i++) {
	 if(e->frame.retcode[i] != frame->retcode[i]) {
	   printk("replay_rt_sigframe: retcode[%d] differs: %lx %lx\n",
	   		i, 
			frame->retcode[i], 
			e->frame.retcode[i]);
	    pause=1;
	 }
	}

	for(i=0; i<_NSIG_WORDS; i++) {
	 if(e->frame.uc.uc_sigmask.sig[i] != frame->uc.uc_sigmask.sig[i]) {
	   printk("replay_rt_sigframe: uc.uc_sigmask.sig[%d] differs: %lx %lx\n",
	   		i, 
			frame->uc.uc_sigmask.sig[i], 
			e->frame.uc.uc_sigmask.sig[i]);
	    pause=1;
	 }
	}

	DIFF(fpstate.cw, "fpstate.cw");
	DIFF(fpstate.sw, "fpstate.sw");
	DIFF(fpstate.tag, "fpstate.tag");
	DIFF(fpstate.ipoff, "fpstate.ipoff");
	DIFF(fpstate.cssel, "fpstate.cssel");
	DIFF(fpstate.dataoff, "fpstate.dataoff");
	DIFF(fpstate.datasel, "fpstate.datasel");
	DIFF(fpstate.status, "fpstate.status");
	DIFF(fpstate.magic, "fpstate.magic");
	DIFF(fpstate.mxcsr, "fpstate.mxcsr");
	DIFF(fpstate.reserved, "fpstate.reserved");

#else
#define DIFF(x, y) \
	if(e->x != frame->x) { 				 \
	  printk("replay_rt_sigframe: %s differs: %lx %lx\n", \
	  	y, frame->x, e->x);		  	 \
  	  pause = 1;					 \
	}
#endif
	DIFF(sig, "sig");
#undef DIFF

     	if(pause) {
     	  printk("   pausing...\n");
	  ztrace_debug_pause();
     	}
#ifdef REPLAY_WHOLE_SIGFRAME
	return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
#else
	frame->uc.uc_mcontext.trapno = e->trapno;
	return __copy_to_user(user_frame, frame, sizeof(*frame));
#endif
}

/*
 * Move this to another file
 */

//#define REPLAY_DEBUG

extern long sys_call_table[];

extern struct ztrace_rununtil_struct zt_rununtil;

#define REPLAY_EXCLUDE_SOME

#ifndef REPLAY_ALL_SYSCALLS
# define REPLAY_UNLOGGED \
	case __NR_sigprocmask: \
	case __NR_mmap: \
	case __NR_rt_sigreturn: \
	case __NR_getpid: \
	case __NR_munmap: \
	case __NR_lseek: \
	case __NR_sigaltstack: \
	case __NR_write: 	\
	case __NR_getguestpgd: \
	case __NR_switchguest: \
	case __NR_rt_sigsuspend: \
	case __NR__llseek:
#else
# define REPLAY_UNLOGGED
#endif

#define REPLAY_LOGGED \
	case __NR_read: \
	case __NR_socketcall: \
	case __NR_gettimeofday:

/*
 * Called before a syscall either during logging or replay; turns on
 * XUSER and/or DATASQUASH if appropriate
 */
asmlinkage void revirt_presys(struct pt_regs regs) {
	struct replay_ctrdata cnt;
	current->current_syscall = regs.eax = regs.orig_eax;
      	copy_branchdata(current, &cnt);
#ifdef REPLAY_DEBUG
	if(current->replay_flags & REPLAY_REPLAY) {
	    printk("revirt_presys: pid %d @ 0x%lx calls syscall %d(0x%lx 0x%lx 0x%lx 0x%lx 0x%lx) [%llx %x %x %x]\n",
			current->pid, regs.eip, current->current_syscall, 
			regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi,
			cnt.hw_branch, cnt.hw_intr, cnt.ex_pf, cnt.ex_dev);
	}
#endif
	if((zt_flags & ZTRACE_BKPT) && (ADJ_BC(cnt) > zt_rununtil.count)) {
		printk("revirt_presys: count %llx, zr wanted %llx\n",
				ADJ_BC(cnt), zt_rununtil.count);
		if(replay_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
			ztrace_debug_pause();
		}
	}
	if((zt_flags & ZTRACE_PERF) && (cnt.hw_branch > 0x100)) {
		printk("revirt_presys: ZT_PEF, hw_count %llx\n",
				cnt.hw_branch);
		if(replay_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
			ztrace_debug_pause();
		}
	}
	if((replay_options.replayflags & REPLAY_OPT_RESTRICT_SYSCALLS)
		&& current->umlinux.umlTraced) {
		if(!FD_ISSET(current->current_syscall, &replay_options.allowed_syscalls)) {
			printk("replay_presys: pid %d @%lx makes forbidden syscall %d(0x%lx 0x%lx 0x%lx 0x%lx 0x%lx)\n",
				current->pid, regs.eip, current->current_syscall,
				regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi);
			if(replay_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
				ztrace_debug_pause();
			}
		}
	}
	switch(current->current_syscall) {
#ifdef REPLAY_EXCLUDE_SOME
	    REPLAY_UNLOGGED
	    case __NR_lexecve:
	    	break;
	    default:
#else
	    REPLAY_LOGGED
#endif
	    	if(current->replay_flags & REPLAY_LOG) 
			current->replay_flags |= REPLAY_XUSER;
		else if(current->replay_flags & REPLAY_REPLAY) {
			current->replay_flags |= REPLAY_DATASQUASH;	
			current->replay_flags |= REPLAY_SYSCALL;
		}
		else {
			printk("revirt_presys: Huh? [pause]\n");
			ztrace_debug_pause();
		}
	}
	if(current->replay_flags & REPLAY_REPLAY)
		if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (current->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
}

asmlinkage void replay_syscall(struct pt_regs * regs);

asmlinkage void revirt_postsys(struct pt_regs regs) {
	struct pt_regs * pregs = &regs;
	current->replay_flags &= ~(REPLAY_DATASQUASH|REPLAY_XUSER|REPLAY_SYSCALL);
	switch(current->current_syscall) {
#ifdef REPLAY_EXCLUDE_SOME
	    REPLAY_UNLOGGED
	    case __NR_lexecve:
	    case __NR_rexecve:
	    	break;
	    default:
#else
	    REPLAY_LOGGED
#endif
		if(current->replay_flags & REPLAY_LOG)
		    log_sysret(pregs);
		else if(current->replay_flags & REPLAY_REPLAY) {
		    replay_syscall(pregs);
		}
		else {
			printk("revirt_postsys: Huh? [pause]\n");
			ztrace_debug_pause();
		}
	}
}

asmlinkage void replay_syscall(struct pt_regs * regs) {
	int ret = -ENOSYS, r=0;
	unsigned int t;
	struct rlog_header rh;
	struct rlog_put_user_entry * pe;
	struct rlog_copy_user_entry * ce;
	struct rlog_entry_sysret * se;
	struct rlog_clear_user_entry *cle;

#ifdef REPLAY_DEBUG
	printk("replay_syscall: rexec returned %d.\n", regs->eax);
#endif
	ztrace_tsk_rexec_done();
	current->nrxuser=0;

#ifdef REPLAY_DEBUG
	printk("replay_syscall: replaying\n");
#endif
	while(1) {
	    demon_read_log(&rh);
	    switch(rh.type) {
	        case ENTRY_SENDERIP:
			break;
	    	case ENTRY_PUT_USER:
		    pe=(struct rlog_put_user_entry *)replay_scall_buf;
		    if(pe->pid != current->pid) {
		    	printk("replay_syscall %d (%u):pid %d, expecting %d\n", 
				current->current_syscall, 
				rh.id, current->pid, pe->pid);
			printk("  pausing...\n");
			ztrace_debug_pause();
		    }
		    /*
		     * Ok, this is wierd.  put_user() is called with 
		     * any size integer, without being told its size, and
		     * is expected to figure it out with compiler techniques.
		     * It does so using the sizeof(*(ptr)) compiler operator.
		     * As a result, __put_user_size is always "called" with
		     * a constant: 1, 2, or 4 (for this architecture).  This
		     * is replaced with switch(2) { } (or 1 or 4), and the 
		     * optimizer optimizes away all other cases, including
		     * the default.
		     *
		     * If ever anyone calls put_user with a pointer not of
		     * size 1,2, or 4, the default: case is instantiated, 
		     * which calls __put_user_bad.  Since there is no
		     * __put_user_bad, this causes the link to fail.
		     *
		     * But it will do the same thing if we call __put_user_size
		     * with a variable, such as pe->ssize, since the compiler
		     * doesn't know what values pe->ssize may take.  So
		     * we call __put_user_size with constants.
		     */

		    if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (current->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
#ifdef REPLAY_DEBUG
		    printk("replay_user: pid %d(%u)__put_user_size(%lx, %lx, %d)",
		    		current->pid, rh.id, pe->xx, pe->pptr, pe->ssize);
#endif
		    switch(pe->ssize) {
		      case 1:
		    	__put_user_size(pe->xx, pe->pptr, 1, r);
			break;
		      case 2:
		    	__put_user_size(pe->xx, pe->pptr, 2, r);
			break;
		      case 4:
		    	__put_user_size(pe->xx, pe->pptr, 4, r);
			break;
		      case 8:
		    	__put_user_size(pe->xx, pe->pptr, 8, r);
			break;
		      default:
		      	printk("replay_user: bad put_user size: %d [pause]\n", 
					pe->ssize);
			ztrace_debug_pause();
		    }
		    current->nrxuser++;
		    if(r!=pe->rretval) {
		    	printk("replay_syscall %d:putuser ret %d,expecting %d\n", 
				current->current_syscall, r, pe->rretval);
		    }
		    break;
		case ENTRY_COPY_USER:
		  do {
		    char * d;
		    int i, l;

		    ce=(struct rlog_copy_user_entry *)replay_scall_buf;
		    if(ce->pid != current->pid) {
		    	printk("replay_syscall %d: pid %d, expecting %d(%u)\n", 
				current->current_syscall, 
				current->pid, 
				ce->pid, rh.id);
			printk("  pausing...\n");
			ztrace_debug_pause();
		    }
		    if(ce->syscall != current->current_syscall) {
		    	printk("replay_syscall:syscall %d, expecting %d(%u) [pause]\n", 
				current->current_syscall, ce->syscall, rh.id);
			ztrace_debug_pause();
		    }
		    l=ce->ssize;
		    t=ce->tto;
		    i=0;
		    if((d=kmalloc(l, GFP_KERNEL))==NULL) {
		    	printk("replay copy_user: cannot allocate buffer.\n");
			BUG();
		    }
		    memcpy(d, replay_scall_buf+sizeof(*ce), l);
		    /*
		     *  current->nrxuser is incremented at the beginning
		     * of a copy_user, but at the end of this, as it needs
		     * to be to interrupt between xusers.  But when 
		     * interrupting an xuser, the logged nrxuser is one
		     * more, so we compensate.
		     */
		    while(zt_flags & ZTRACE_XUSER_STOP && 
		    	   (zt_xstop.nrxuser-(zt_xstop.edi!=0))
			   			==current->nrxuser) {
			int delta;
			printk("copy_user: edi=0x%x\n", zt_xstop.edi);
			if(!zt_xstop.edi) {
				ztrace_tsk_xuser_schedule();
				continue;
			}
			delta=zt_xstop.edi - t;
			if(delta > l) {
			    printk("rp_copyuser: delta %d, l %d\n", 
			    		delta, l);
			    printk("  pausing...\n");
			    ztrace_debug_pause();
			    BUG();
			}
			r=delta;
			printk("replay syscall: pid %d __copy_user_atom(0x%x,0x%lx,%d)\n",
					current->pid, t, (long unsigned int)d+i, r);
			__copy_user_atom(t,d+i,r);
			if(r) {
		    	 printk("rp_syscall:__copy_user_atom failed at %d\n",
					r);
			 printk("   syscall:%d nrsyscalls:%d\n", ce->syscall,
					ce->nrsyscalls);
			 //BUG();
		    	}
			t+=delta;
			i+=delta;
			l-=delta;
			ztrace_tsk_xuser_schedule();
		    }
		    r=l;
#ifdef REPLAY_DEBUG
		    printk("replay syscall: pid %d(%u) __copy_user_atom(0x%lx,0x%lx,%d)\n",
					current->pid, rh.id, 
					t, (long unsigned int)d+i, r);
#endif
		    __copy_user_atom(t, d+i, r);
		    current->nrxuser++;
		    kfree(d);
		    if(r) {
		    	printk("replay_syscall:__copy_user_atom failed at %d\n",
					r);
			printk("   syscall:%d(%u) nrsyscalls:%d\n", ce->syscall,
					rh.id, ce->nrsyscalls);
			//BUG();
		    }
		  } while(0);
		    break;
		case ENTRY_CLEAR_USER:
		    cle=(struct rlog_clear_user_entry *)replay_scall_buf;
		    if(cle->pid != current->pid) {
		    	printk("replay_syscall %d: pid %d, expecting %d [pause]\n", 
				current->current_syscall, current->pid, 
				cle->pid);
			ztrace_debug_pause();
		    }
		    if(cle->syscall != current->current_syscall) {
		    	printk("replay_syscall:syscall %d, expecting %d [pause]\n", 
				current->current_syscall, cle->syscall);
			ztrace_debug_pause();
		    }
		    if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (current->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
		    r=__clear_user((void *)cle->to, cle->n);
		    current->nrxuser++;
		    if(r) {
		    	printk("replay_syscall:__clear_user failed at %d\n",
					r);
			printk("   syscall:%d \n", cle->syscall);
			//BUG();
		    }
		    break;
		case ENTRY_SYSRET:
		    se=(struct rlog_entry_sysret *)replay_scall_buf;
		    if(se->pid != current->pid) {
		    	printk("replay_syscall sys %d id %u: pid %d, expecting %d [pause]\n", 
				current->current_syscall, 
				rh.id,
				current->pid, 
				se->pid);
			ztrace_debug_pause();
		    }
		    if(se->syscall != current->current_syscall) {
		    	printk("replay_syscall:syscall %d, expecting %d(%u) [pause]\n", 
				current->current_syscall, se->syscall,
				rh.id);
			ztrace_debug_pause();
		    }
		    if((zt_flags & ZTRACE_XUSER_STOP) && 
		       (current->nrxuser==zt_xstop.nrxuser)) 
		       		ztrace_tsk_xuser_schedule();
		    /*
		     *  Check for syscalls with changed values....
		     */
		   if((se->syscall != __NR_read) 
		   	&& (se->syscall != __NR_socketcall)) 
		{
		    int pause=0;
#ifdef REPLAY_WHOLE_PTREGS
		    if(se->regs.ebx != regs->ebx) {
		    	printk("replay_syscall: rex'd ebx differs: %lx %lx\n",
					regs->ebx, se->regs.ebx);
			pause=1;
		    }
		    if(se->regs.ecx != regs->ecx) {
		    	printk("replay_syscall: rex'd ecx differs: %lx %lx\n",
					regs->ecx, se->regs.ecx);
			pause=1;
		    }
		    if(se->regs.edx != regs->edx) {
		    	printk("replay_syscall: rex'd edx differs: %lx %lx\n",
					regs->edx, se->regs.edx);
			pause=1;
		    }
#else
		    if(se->regs.eax != regs->eax) {
		    	printk("replay_syscall: rex'd eax differs: %lx %x\n",
					regs->eax, se->regs.eax);
			pause=1;
		    }
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
		    do {
		    	struct replay_ctrdata rctr;
			copy_branchdata(current, &rctr);
			if(ADJ_BC(se->ctr) != ADJ_BC_R(rctr)) {
			    printk("replay_syscall: count differs! %llx %llx (%llx) sc %d @ 0x%x\n",
			    	ADJ_BC_R(rctr), ADJ_BC(se->ctr), 
				ADJ_BC(se->ctr) - ADJ_BC_R(rctr),
				current->current_syscall,
				regs->eip);
			printk("replay_syscall: #%d [%llx %x %x] [%llx %x %x]\n",
				rh.id,
			    	rctr.hw_branch,
				rctr.hw_intr,
				rctr.ex_pf,
				se->ctr.hw_branch,
				se->ctr.hw_intr,
				se->ctr.ex_pf);
			    if(replay_options.replayflags & REPLAY_OPT_WARNING_HALTS)
			    	ztrace_debug_pause();
			}
		    } while(0);
#endif
		    if(pause){
		    	printk("replay_syscall: syscall %d(%u) should probably be logged [pause]\n",
				se->syscall, rh.id);
			ztrace_debug_pause();
		    }
		   }
#if 0
		    /*
		     * sigreturn should change registers... maybe eventually
		     * check for execve as well?
		     */
		    if(se->syscall != __NR_rt_sigreturn) {
		     int pause=0;
		     if(rexec_ret != se->regs.eax && se->syscall != __NR_read) {
		     	printk("replay_syscall: eax differs: %lx %lx\n",
					r, se->regs.eax);
			pause=1;
		     }
		     if(regs->ebx != se->regs.ebx) {
		     	printk("replay_syscall: ebx differs: %lx %lx\n",
					regs->ebx, se->regs.ebx);
			pause=1;
		     }
		     if(regs->ecx != se->regs.ecx) {
		     	printk("replay_syscall: ecx differs: %lx %lx\n",
					regs.ecx, se->regs.ecx);
			pause=1;
		     }
		     if(regs.edx != se->regs.edx) {
		     	printk("replay_syscall: edx differs: %lx %lx\n",
					regs.edx, se->regs.edx);
			pause=1;
		     }
		     if(regs.esi != se->regs.esi) {
		     	printk("replay_syscall: esi differs: %lx %lx\n",
					regs.esi, se->regs.esi);
			pause=1;
		     }
		     if(regs.edi != se->regs.edi) {
		     	printk("replay_syscall: edi differs: %lx %lx\n",
					regs.edi, se->regs.edi);
			pause=1;
		     }
		     /*
		      * because [rt_]sigsuspend calls do_signal directly,
		      * do_signal's changes to the esp don't happen.  Let
		      * it return them to what they would have been...
		      */
		     if(regs.esp != se->regs.esp
		     	  && !(se->syscall == __NR_sigsuspend 
			  	|| se->syscall == __NR_rt_sigsuspend)) {
		     	printk("replay_syscall: esp differs: %lx %lx\n",
					regs.esp, se->regs.esp);
			pause=1; 
		     }
		     if(regs.ebp != se->regs.ebp) {
		     	printk("replay_syscall: ebp differs: %lx %lx\n",
					regs.ebp, se->regs.ebp);
			pause=1;
		     }
		     if(pause) {
		     	printk("   pausing...\n");
			ztrace_debug_pause();
		     }
		    }
#endif

//#ifdef REPLAY_WHOLE_PTREGS
#if 0
		    *regs=se->regs;
#else
		    regs->eax = se->regs.eax;
#endif
		    ret = regs->eax;
#ifdef REPLAY_DEBUG
		    printk("replay_sycall: pid %d ret %d @ 0x%lx\n", 
		    		current->pid, ret, regs->eip);
# ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
		    printk("replay_syscall: logged count: [%llx %x %x %x]\n",
		    		se->ctr.hw_branch,
		    		se->ctr.hw_intr,
		    		se->ctr.ex_pf,
		    		se->ctr.ex_dev);
# endif
#endif
		    current->nrxuser++;
		    goto out;
		default:
		    printk("Bad syscall entry: %d (%u) [pause]\n", 
				    rh.type, rh.id);
		    if(rh.type == ENTRY_RT_SIGFRAME) {
		    	struct rlog_entry_rt_sigframe * sys;
		    	sys=(struct rlog_entry_rt_sigframe *)replay_scall_buf;
#ifdef REPLAY_WHOLE_SIGFRAME
			printk(" sig: %d eip:%lx\n", sys->frame.sig,
					sys->frame.uc.uc_mcontext.eip);
#else
			printk(" sig: %d\n", sys->sig);
#endif
		    }
		    ztrace_debug_pause();
	    }
	}
    out:
	return;
}

unsigned long replay_copyuser(unsigned long to, unsigned long from, 
	unsigned long size) {
    struct rlog_copy_user_entry le;

    /*
     * There are times when the same function may write either to userland
     * or to kernel; e.g., low-level file reads.  We need to make sure to
     * log & replay only those which go into userland.
     *
     * FIXME: we need to use a #define for this to make it portable
     */
    if(to > 0xc0000000) {
    	int flag;
	flag = current->replay_flags & REPLAY_XUSER;
	current->replay_flags &= ~REPLAY_XUSER;
	__copy_user_atom(to, from, size);
	current->replay_flags |= flag;
	return size;
    }

    if(current->replay_flags & REPLAY_LOG) {
	if( 0 /*current->replay_flags & REPLAY_DATAPASS */) {
	    /* log size, no data */
	    le.type=ENTRY_COPY_USER_NODATA;
	    le.tto=to;
	    le.ssize=size;
	    le.pid=current->pid;
	    le.nrsyscalls=current->nrsyscalls;
	    le.syscall=current->current_syscall;
	    demon_write_log((char *)&le, le.type, sizeof(le));
	} else if(current->replay_flags & REPLAY_XUSER) {
	    /* log both metadata and data */
            struct iovec v[2]; 
            le.type = ENTRY_COPY_USER;
            le.tto = to;
            le.ssize = size;
            le.pid = current->pid;
            le.nrsyscalls = current->nrsyscalls;
            le.syscall = current->current_syscall;
            v[0].iov_base = &le;
            v[0].iov_len = sizeof(le);
            v[1].iov_base = (void *)from;
            v[1].iov_len = size;
            demon_writev_log(le.type, v, 2);
	}
	__copy_user_atom(to, from, size);
	return size;
    }
    if(current->replay_flags & REPLAY_REPLAY) {
	if(current->replay_flags & REPLAY_DATASQUASH) {
	    if(to < 0xc0000000) {
#ifdef REPLAY_DEBUG
		printk("squashing copy_user(0x%lx,0x%lx,%ld)\n",
			to, from, size);
#endif
		return 0;
	    } else {
	    	printk("WARNING: calling copy_user with kernel destination\n");
		printk("   copy_user(0x%lx,0x%lx,%ld)\n", to, from, size);
		ztrace_debug_pause();
		BUG();
	    }
	}
	else {
		__copy_user_atom(to, from, size);
		return size;
	}
    }
    printk("replay_copy_user: huh?\n");
    ztrace_debug_pause();
    BUG();
    return -1;
}

void replay_exit(struct task_struct * tsk) {
        if (tsk->replay_flags&REPLAY_LOG)  {
                log_exit();
        } else if(tsk->replay_flags & REPLAY_REPLAY) {
	    if(tsk->sched) {
                if(zt_flags & ZTRACE_WAITEXIT) {
                        printk("do_exit: pid %d waking up scheduler\n",
                                tsk->pid);
                        /* FIXME: do we need to worry about locks here? */
                        tsk->sched->replay_wait &= ~REPLAY_WAIT;
                        zt_flags &= ~ZTRACE_WAITEXIT;
                } else {
                        printk("do_exit: pid %d exiting unexpectedly?\n",
                                tsk->pid);
                        tsk->sched=NULL;
                        BUG();
                }
	    }
	    /*
	     * FIXME -- assumes only one proc for now, just for UML.  If we
	     * go back to host-based replay, we need to keep the file open.
	     */
	    demon_read_close();
        }
}
