/*
 * revirt/wraw.c
 *
 * Hacked from linux/drivers/char/raw.c
 *
 * Front-end raw character devices.  These can be bound to any block
 * devices to provide genuine Unix raw character device semantics.
 *
 * We reserve minor number 0 for a control interface.  ioctl()s on this
 * device are used to bind the other minor numbers to block devices.
 */

#include <linux/fs.h>
#include <linux/iobuf.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/raw.h>
#include <linux/capability.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include <revirt/wraw.h>

#define dprintk(x...) 

typedef struct wraw_device_data_s {
	struct block_device *binding;
	int inuse, sector_size, sector_bits;
	struct semaphore mutex;
	struct wraw_watch_s * watch;
} wraw_device_data_t;

//static struct wraw_watch_s watch;
//static struct wraw_write_record wr_buf[WBUFSIZE];
//static char __data_buf[DBUFSIZE+4096];
//static char *__data_buf=NULL;
//static char * data_buf=NULL;

static wraw_device_data_t wraw_devices[256];

static ssize_t rw_wraw_dev(int rw, struct file *, char *, size_t, loff_t *);

ssize_t	wraw_read(struct file *, char *, size_t, loff_t *);
ssize_t	wraw_write(struct file *, const char *, size_t, loff_t *);
int	wraw_open(struct inode *, struct file *);
int	wraw_release(struct inode *, struct file *);
int	wraw_ctl_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int	wraw_ioctl(struct inode *, struct file *, unsigned int, unsigned long);


static struct file_operations wraw_fops = {
	read:		wraw_read,
	write:		wraw_write,
	open:		wraw_open,
	release:	wraw_release,
	ioctl:		wraw_ioctl,
};

static struct file_operations wraw_ctl_fops = {
	ioctl:		wraw_ctl_ioctl,
	open:		wraw_open,
};

int init_watchstructure(wraw_device_data_t * wdd) {
	struct wraw_watch_s * watch;

	dprintk("Requesting %d bytes of GFP_KERNEL\n", 
			sizeof(struct wraw_watch_s));
	watch=kmalloc( sizeof(struct wraw_watch_s), GFP_KERNEL);
	dprintk("wraw: kmalloc returned 0x%x\n", watch);
	if(watch == NULL) {
		return -ENOMEM;
	}
	printk("wraw: initializing stuff\n");

	watch->watched=0;
	watch->tsk=NULL;
	init_MUTEX(&watch->mutex);
	init_waitqueue_head(&watch->wait_read);
	init_waitqueue_head(&watch->wait_write);
	watch->datamap_u=NULL;
	watch->maxoffset=0;
	watch->maxindex=0;
	//watch->datamap_k=NULL;
	watch->written_u=NULL;
	watch->maxwr=watch->curwr=watch->err=0;
	watch->databuf_u=NULL;

	dprintk("wraw: about to allocate kiovecs\n");
	if(alloc_kiovec(WBUFSIZE, watch->iovec_page)<0) {
		printk("wraw: alloc_kiovec failed\n");
		watch->maxwr=-1;
	}
	dprintk("Setting wdd->watch to watch\n");
	wdd->watch = watch;
	return 0;
}

static int __init wraw_init(void)
{
	int i;
	register_chrdev(WRAW_MAJOR, "wraw", &wraw_fops);

	for (i = 0; i < 256; i++)  {
		wraw_devices[i].binding=NULL;
		wraw_devices[i].inuse=0;
		init_MUTEX(&wraw_devices[i].mutex);
		wraw_devices[i].watch=NULL;
	}

	return 0;
}

__initcall(wraw_init);

static loff_t get_device_size(int minor) {
	loff_t limit;
	kdev_t dev;

	dev = to_kdev_t(wraw_devices[minor].binding->bd_dev);

	if (blk_size[MAJOR(dev)])
		limit = (((loff_t) blk_size[MAJOR(dev)][MINOR(dev)]) << BLOCK_SIZE_BITS) ;
        else
		limit = INT_MAX;
	return limit;
}


int wraw_ioctl(struct inode *inode, 
		struct file *filp, 
		unsigned int command, 
		unsigned long argp) {
    //struct wraw_watch_s * wp;
    struct wraw_grab_arg_s grab_arg;
    struct wraw_listen_arg_s listen_arg;
    struct wraw_watch_s * watch;
    int minor, err, i;

    minor = MINOR(filp->f_dentry->d_inode->i_rdev);
    err=-EINVAL;
    //wp=&wraw_devices[minor].watch;
    watch = wraw_devices[minor].watch;

    switch(command) {
    	case WRAW_GRAB:
	    dprintk("wraw: Grabbing general device mutex\n");
	    down(&wraw_devices[minor].mutex);
	    if(watch==NULL) {
	    	dprintk("wraw: watch not initialized, initializing\n");
	    	init_watchstructure(&wraw_devices[minor]);
		watch=wraw_devices[minor].watch;
		if(watch==NULL) {
		   dprintk("wraw: init_watchstructure failed.\n");
		   err=-ENOMEM;
		   up(&wraw_devices[minor].mutex);
		   goto out;
		}
		dprintk("wraw: init_watchstructure succeeded.\n");
	    }
	    up(&wraw_devices[minor].mutex);
	    dprintk("wraw: grabbing watch mutex\n");
	    down(&watch->mutex);
	    if(watch->watched) {
	    	err=-EBUSY;
		goto out_up;
	    }
	    if(copy_from_user(&grab_arg, (void *)argp, sizeof(grab_arg))<0) {
	    	err=-EFAULT;
		goto out_up;
	    }
	    grab_arg.maxoffset=watch->maxoffset=get_device_size(minor);

	    dprintk("wraw: setting maxoffset to %lld\n", watch->maxoffset);

	    if(copy_to_user((void *)argp, &grab_arg, sizeof(grab_arg))<0) {
	    	err=-EFAULT;
		goto out_up;
	    }

	    watch->maxindex = OFF2BYTE(watch->maxoffset);
	    watch->maxwr = watch->curwr=0;

	    watch->watched=1;
	    watch->tsk=current;
	    watch->err=0;


	    err=0;
	    up(&watch->mutex);
	    break;	    
    	case WRAW_LISTEN:
	    if(!watch) {
	    	err=-ENOENT;
		goto out;
	    }
	    down(&watch->mutex);
	    if(!watch->watched || watch->tsk!=current) {
		err=-EINVAL;
		goto out_up;
	    }
	    if(copy_from_user(&listen_arg, (void *)argp, sizeof(listen_arg))) {
	    	err=-EFAULT;
		goto out_up;
	    }

	    watch->datamap_u=listen_arg.datamap_u;

	    err=0;
	    if(listen_arg.written_u) {
		int e=0, sector_bits;
		sector_bits=wraw_devices[minor].sector_bits;
		watch->maxwr= (listen_arg.maxwr < WBUFSIZE ) ? 
	    			listen_arg.maxwr : WBUFSIZE ;
		watch->written_u=listen_arg.written_u;
		watch->databuf_u = listen_arg.databuf_u;
		for(i=0; i<watch->maxwr; i++) {
		    e =map_user_kiobuf(READ, 
		    	watch->iovec_page[i], 
			(unsigned long) (watch->databuf_u + (DBUFGRAIN * i)), 
			DBUFGRAIN);
		    if(e) {
		    	printk("wraw: %d: map_user_kiobuf returns %d\n",
					i, e);
		    	break;
		    }
		}
		if(e) {
		    watch->maxwr=i;
		    if(i==0) {
			err=-ENOMEM;
			watch->maxwr=0;
			goto out_up;
		    }
		}
	    } else {
	    	watch->maxwr=listen_arg.maxwr;
	    }
	    watch->err=0;
	    watch->curwr=0;

	    /*
	     * FIXME -- this is a race condition for SMP or pre-emptible
	     */
	    wake_up_interruptible(&watch->wait_write);
	    while((watch->curwr<watch->maxwr) && (err==0)) {
                up(&watch->mutex);
		/* FIXME -- SMP / preemptible kernel race here */
                err=wait_event_interruptible(watch->wait_read, 
			watch->curwr >= watch->maxwr);
                down(&watch->mutex); 
	    }
	    if(err==-ERESTARTSYS)
		    err=-EINTR;
	    /*
	     * Unmap all the iobufs that weren't used; others are
	     * unmapped as they're used
	     */
	    if(listen_arg.written_u) {
		for(i=watch->curwr; i<watch->maxwr; i++) {
		    unmap_kiobuf(watch->iovec_page[i]);
		}
	    }
	    listen_arg.curwr=watch->curwr;
	    if(copy_to_user((void *)argp, &listen_arg, sizeof(listen_arg))) {
		    err=-EFAULT;
		    goto done;
	    }
	   done:

	    watch->maxwr=watch->curwr=0;
	    break;
	case WRAW_RELEASE:
	    if(!watch) {
	    	err=-ENOENT;
		goto out;
	    }
	    down(&watch->mutex);
	    watch->watched=0;
	    watch->datamap_u=NULL;
	    watch->written_u=NULL;
	    watch->databuf_u=NULL;
	    watch->maxwr=0;
	    watch->curwr=0;
	    wake_up_interruptible(&watch->wait_read);
	    wake_up_interruptible(&watch->wait_write);
	    err=0;
	    break;
	default:
	    err=-EINVAL;
	    goto out;
    }

    out_up:
    	up(&watch->mutex);

    out:
       return err;
}

/* 
 * Open/close code for raw IO.
 */

int wraw_open(struct inode *inode, struct file *filp)
{
	int minor;
	struct block_device * bdev;
	kdev_t rdev;	/* it should eventually go away */
	int err;
	int sector_size;
	int sector_bits;

	minor = MINOR(inode->i_rdev);
	
	/* 
	 * Is it the control device? 
	 */
	
	if (minor == 0) {
		filp->f_op = &wraw_ctl_fops;
		return 0;
	}
	
	if (!filp->f_iobuf) {
		err = alloc_kiovec(1, &filp->f_iobuf);
		if (err)
			return err;
	}

	down(&wraw_devices[minor].mutex);
	/*
	 * No, it is a normal wraw device.  All we need to do on open is
	 * to check that the device is bound, and force the underlying
	 * block device to a sector-size blocksize. 
	 */

	bdev = wraw_devices[minor].binding;
	err = -ENODEV;
	if (!bdev)
		goto out;

	atomic_inc(&bdev->bd_count);
	rdev = to_kdev_t(bdev->bd_dev);
	err = blkdev_get(bdev, filp->f_mode, 0, BDEV_RAW);
	if (err)
		goto out;
	
	/*
	 * Don't change the blocksize if we already have users using
	 * this device 
	 */

	if (wraw_devices[minor].inuse++)
		goto out;

	/* 
	 * Don't interfere with mounted devices: we cannot safely set
	 * the blocksize on a device which is already mounted.  
	 */
	
	sector_size = 512;
	if (is_mounted(rdev)) {
		if (blksize_size[MAJOR(rdev)])
			sector_size = blksize_size[MAJOR(rdev)][MINOR(rdev)];
	} else {
		if (hardsect_size[MAJOR(rdev)])
			sector_size = hardsect_size[MAJOR(rdev)][MINOR(rdev)];
	}

	set_blocksize(rdev, sector_size);
	wraw_devices[minor].sector_size = sector_size;

	for (sector_bits = 0; !(sector_size & 1); )
		sector_size>>=1, sector_bits++;
	wraw_devices[minor].sector_bits = sector_bits;

 out:
	up(&wraw_devices[minor].mutex);
	
	return err;
}

int wraw_release(struct inode *inode, struct file *filp)
{
	int minor;
	struct block_device *bdev;
	
	minor = MINOR(inode->i_rdev);
	down(&wraw_devices[minor].mutex);
	bdev = wraw_devices[minor].binding;
	wraw_devices[minor].inuse--;
	up(&wraw_devices[minor].mutex);
	blkdev_put(bdev, BDEV_RAW);
	return 0;
}



/*
 * Deal with ioctls against the raw-device control interface, to bind
 * and unbind other raw devices.  
 */

int wraw_ctl_ioctl(struct inode *inode, 
		  struct file *flip,
		  unsigned int command, 
		  unsigned long arg)
{
	struct raw_config_request rq;
	int err = 0;
	int minor;
	
	switch (command) {
	case RAW_SETBIND:
	case RAW_GETBIND:

		/* First, find out which raw minor we want */

		err = copy_from_user(&rq, (void *) arg, sizeof(rq));
		if (err)
			break;
		
		minor = rq.raw_minor;
		if (minor <= 0 || minor > MINORMASK) {
			err = -EINVAL;
			break;
		}

		if (command == RAW_SETBIND) {
			/*
			 * This is like making block devices, so demand the
			 * same capability
			 */
			if (!capable(CAP_SYS_ADMIN)) {
				err = -EPERM;
				break;
			}

			/* 
			 * For now, we don't need to check that the underlying
			 * block device is present or not: we can do that when
			 * the raw device is opened.  Just check that the
			 * major/minor numbers make sense. 
			 */

			if ((rq.block_major == NODEV && 
			     rq.block_minor != NODEV) ||
			    rq.block_major > MAX_BLKDEV ||
			    rq.block_minor > MINORMASK) {
				err = -EINVAL;
				break;
			}
			
			down(&wraw_devices[minor].mutex);
			if (wraw_devices[minor].inuse) {
				up(&wraw_devices[minor].mutex);
				err = -EBUSY;
				break;
			}
			if (wraw_devices[minor].binding)
				bdput(wraw_devices[minor].binding);
			wraw_devices[minor].binding = 
				bdget(kdev_t_to_nr(MKDEV(rq.block_major, rq.block_minor)));
			up(&wraw_devices[minor].mutex);
		} else {
			struct block_device *bdev;
			kdev_t dev;

			bdev = wraw_devices[minor].binding;
			if (bdev) {
				dev = to_kdev_t(bdev->bd_dev);
				rq.block_major = MAJOR(dev);
				rq.block_minor = MINOR(dev);
			} else {
				rq.block_major = rq.block_minor = 0;
			}
			err = copy_to_user((void *) arg, &rq, sizeof(rq));
		}
		break;
		
	default:
		err = -EINVAL;
	}
	
	return err;
}

static int wait_for_space(struct wraw_watch_s * watch) {
    int intr;
    while(watch->curwr >= watch->maxwr 
		    && watch->watched) {
    	/*
	 * FIXME -- this is a race in SMP or pre-emptible kern
	 */
	dprintk("curwr: %d maxwr: %d, watched %d; sleeping\n",
		watch->curwr, watch->maxwr, watch->watched);
    	up(&watch->mutex);
	intr=wait_event_interruptible(watch->wait_write,
		(watch->curwr < watch->maxwr) || !watch->watched);
	down(&watch->mutex);
	if(intr || watch->watched==0)
	    goto wout; 
    }
    return 0;
   wout:
    return -1;
}

static int check_datamap(struct wraw_watch_s * watch,
		kdev_t dev, loff_t old_off, size_t size, 
		int sector_bits, int sector_size, unsigned long wlimit) {
    unsigned long byte;
    loff_t end;
    int i, j, k, B_start, B_end, b_start, b_end, err;

    dprintk("Task %d attempting to grab watch->mutex\n",
   		current->pid);

    down(&watch->mutex);
    end=old_off + size - 1;

    B_start=OFF2BYTE(old_off);
    b_start=OFF2BIT(old_off);
    B_end=OFF2BYTE(end);
    b_end=OFF2BIT(end);

    /* 
     * Go through all bytes in this range, from B_start
     * to B_end
     */
    dprintk("Transferred from %llx to %llx, looking for bits\n",
    		old_off, end);
    for(i=B_start; i<=B_end; i++) {
    	int mod=0;
    	if(access_process_vm(watch->tsk,
		(unsigned int)(watch->datamap_u+i),
		&byte, sizeof(byte), 
		ACCESS_RD) < sizeof(byte)){
	     printk("wraw: datamap_u read failed: 0x%x\n",
	     		(unsigned int)(watch->datamap_u+i));    
	    break;
	}
	dprintk("Read 0x%x from datamap_u[%d]\n", 
		byte, i);
	/*
	 * For each byte, check appropriate bit ranges.  
	 *
	 * If this is the first byte (i == B_start) only
	 * start at the b_start'th bit; otherwise, start
	 * at bit 0.
	 *
	 * If this is the last byte (i == B_end) only go
	 * through the b_end'th bit; otherwise, go to the
	 * end of the byte (31)
	 *
	 * Note that this works even if B_start==B_end.
	 */
	for(j = ((i==B_start) ? b_start : 0) ;
	    j <= ((i==B_end) ? b_end : 31);
	    j++) {
	    if( (byte & (1 << j)) == 0) {
	    	/*
		 * Read data, record page
		 * Make sure to account for non-multiples
		 * of 4096.
		 */
		if(watch->written_u && watch->err==0) {
		    loff_t woff;
		    int wstart, wblocks;

		    if(wait_for_space(watch) <0) {
		       	goto loop_out;
		    }

		    woff=BB2OFF(i, j);
		    wstart = woff >> sector_bits;
		    wblocks = DBUFGRAIN >> sector_bits;
		    //wlimit = limit >> sector_bits;

		   /*
		    * Fill the block thing up.
		    * We might not actually be on a page boundary, so if
		    * a block is past the end of the device, fill the page
		    * with stuff.  This may be inefficient, I'm not sure;
		    * but it only happens for the very last page, which should
		    * only be written at most once per undo window; correct
		    * and slow but very uncommon, hopefully.
		    */
		   dprintk("%llx -> (%d, %d) -> %llx -> (%d, %d) -> w[%d]\n",
		   	old_off, i, j, woff, wstart, wblocks, watch->curwr);
		   if(wstart + wblocks >= wlimit) {
		   	printk("wraw: Warning copying last block\n");
		   }
		   for(k=0; k<wblocks; k++) {
		    watch->iovec_page[watch->curwr]->blocks[k] 
		    	= (wstart+k < wlimit) ? wstart+k : wlimit;
		   }
		   err = brw_kiovec(READ, 1, 
		   		&watch->iovec_page[watch->curwr], 
		   		dev, 
				watch->iovec_page[watch->curwr]->blocks, 
				sector_size);

		    if (err > 0) {
			mark_dirty_kiobuf(watch->iovec_page[watch->curwr], err);
			if(access_process_vm(watch->tsk,
			    (unsigned int) (watch->written_u+watch->curwr),
		    	    &woff, sizeof(woff), ACCESS_WR) < sizeof(woff)) {
    
    			    printk("wraw: written_u write failed: 0x%x\n",
  				(unsigned int) (watch->written_u + watch->curwr));
    			    watch->err=-EFAULT;
       			}
		    }
		    else {
		    	printk("wraw: COW returned %d\n", err);
			watch->err=err;
		    }
		    unmap_kiobuf(watch->iovec_page[watch->curwr]);
		    
		}
	    	watch->curwr++;
		if(watch->curwr >= watch->maxwr)
		    wake_up_interruptible(&watch->wait_read);
		byte |= (1 << j);
		mod=1;
	    }
	}
	if(mod) {
	 if(access_process_vm(watch->tsk,
		(unsigned int)(watch->datamap_u+i),
		&byte, sizeof(byte), 
		ACCESS_WR) < sizeof(byte)){
	     printk("wraw: datamap_u write failed: 0x%x\n",
	     		(unsigned int)(watch->datamap_u+i));
	    break;
	 }
	 dprintk("Wrote 0x%x to datamap_u[%d]\n",
		byte, i);
	}

    }
   loop_out:
    dprintk("Bit-looking done; curwr %d, maxwr %d\n",
    		watch->curwr, watch->maxwr);
    dprintk("Unlocking watch->mutex\n");
    up(&watch->mutex);

   return 0;
}

ssize_t	wraw_read(struct file *filp, char * buf, 
		 size_t size, loff_t *offp)
{
	return rw_wraw_dev(READ, filp, buf, size, offp);
}

ssize_t	wraw_write(struct file *filp, const char *buf, 
		  size_t size, loff_t *offp)
{
	return rw_wraw_dev(WRITE, filp, (char *) buf, size, offp);
}

#define SECTOR_BITS 9
#define SECTOR_SIZE (1U << SECTOR_BITS)
#define SECTOR_MASK (SECTOR_SIZE - 1)

ssize_t	rw_wraw_dev(int rw, struct file *filp, char *buf, 
		   size_t size, loff_t *offp)
{
	struct kiobuf * iobuf;
	int		new_iobuf;
	int		err = 0;
	unsigned long	blocknr, blocks;
	size_t		transferred=0;
	int		iosize;
	int		i;
	int		minor;
	kdev_t		dev;
	unsigned long	limit;

	int		sector_size, sector_bits, sector_mask;
	int		max_sectors;
	struct wraw_watch_s * watch;
	
	/*
	 * First, a few checks on device size limits 
	 */

	minor = MINOR(filp->f_dentry->d_inode->i_rdev);

	watch = wraw_devices[minor].watch;

	new_iobuf = 0;
	/*
	 * By moving iobuf lower, we can block in copy-on-write without
	 * bothering about the f_iobuf_lock; furthermore, we don't have
	 * to goto out_free in many cases.
	 */


	dev = to_kdev_t(wraw_devices[minor].binding->bd_dev);
	sector_size = wraw_devices[minor].sector_size;
	sector_bits = wraw_devices[minor].sector_bits;
	sector_mask = sector_size- 1;
	max_sectors = KIO_MAX_SECTORS >> (sector_bits - 9);
	
	if (blk_size[MAJOR(dev)])
		limit = (((loff_t) blk_size[MAJOR(dev)][MINOR(dev)]) << BLOCK_SIZE_BITS) >> sector_bits;
	else
		limit = INT_MAX;
	dprintk ("rw_wraw_dev: dev %d:%d (+%d)\n",
		 MAJOR(dev), MINOR(dev), limit);
	
	err = -EINVAL;
	if ((*offp & sector_mask) || (size & sector_mask))
		goto out; /* was out_free */
	err = 0;
	if (size)
		err = -ENXIO;
	if ((*offp >> sector_bits) >= limit)
		goto out; /* was out_free */

	/*
	 * GWD -- copy-on-write functonality: Check to see if we
	 * need to copy anything; if so, do it.
	 */
	if(watch && watch->watched && watch->err==0 && rw==WRITE) {
	    size_t lsize = size;
	    if(*offp > watch->maxoffset) {
	    	printk("wraw: Warning, 0x%llx > 0x%llx, ignoring\n",
			*offp, watch->maxoffset);

	    } else {
	     if((*offp + lsize > watch->maxoffset)) {
	    	lsize = watch->maxoffset - *offp;
		printk("wraw: Warining size %d exceeds limit, setting to %d\n",
			size, lsize);
	     }
	     check_datamap(watch, dev, *offp, lsize, sector_bits, 
	     			sector_size, limit);
	    }
	}

	iobuf = filp->f_iobuf;
	if (test_and_set_bit(0, &filp->f_iobuf_lock)) {
		/*
		 * A parallel read/write is using the preallocated iobuf
		 * so just run slow and allocate a new one.
		 */
		err = alloc_kiovec(1, &iobuf);
		if (err)
			goto out;
		new_iobuf = 1;
	}

	/*
	 * Split the IO into KIO_MAX_SECTORS chunks, mapping and
	 * unmapping the single kiobuf as we go to perform each chunk of
	 * IO.  
	 */
	transferred = 0;
	blocknr = *offp >> sector_bits;
	while (size > 0) {
		blocks = size >> sector_bits;
		if (blocks > max_sectors)
			blocks = max_sectors;
		if (blocks > limit - blocknr)
			blocks = limit - blocknr;
		if (!blocks)
			break;

		iosize = blocks << sector_bits;

		err = map_user_kiobuf(rw, iobuf, (unsigned long) buf, iosize);
		if (err)
			break;

		for (i=0; i < blocks; i++) 
			iobuf->blocks[i] = blocknr++;

		err = brw_kiovec(rw, 1, &iobuf, dev, iobuf->blocks, sector_size);

		if (rw == READ && err > 0)
			mark_dirty_kiobuf(iobuf, err);
		
		if (err >= 0) {
			transferred += err;
			size -= err;
			buf += err;
		}

		unmap_kiobuf(iobuf);

		if (err != iosize)
			break;
	}
	
	if (transferred) {
		*offp += transferred;
		err = transferred;
	}

	if (!new_iobuf)
		clear_bit(0, &filp->f_iobuf_lock);
	else
		free_kiovec(1, &iobuf);

 out:	
	return err;
}
