/*
 *  linux/drivers/video/umlcon.c -- Low level UMLinux based console driver
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/console_struct.h>
#include <linux/string.h>
#include <linux/kd.h>
#include <linux/slab.h>
#include <linux/vt_kern.h>
#include <linux/selection.h>
#include <linux/spinlock.h>
#include <linux/ioport.h>
#include <linux/init.h>

#include <linux/assert.h>
#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>


static spinlock_t uml_lock = SPIN_LOCK_UNLOCKED;

/*
 *  Interface used by the world
 */

static const char *umlcon_startup(void);
static void umlcon_init(struct vc_data *c, int init);
static void umlcon_deinit(struct vc_data *c);
static void umlcon_cursor(struct vc_data *c, int mode);
static void umlcon_clear(struct vc_data *c, int sy, int sx, int height,
		int width);
static void umlcon_putc(struct vc_data *c, int ch, int ypos, int xpos);
static void umlcon_putcs(struct vc_data *c, const unsigned short *s,
		int count, int ypos, int xpos);
static int umlcon_switch(struct vc_data *c);
static int umlcon_blank(struct vc_data *c, int blank);
static int umlcon_font_op(struct vc_data *c, struct console_font_op *op);
static int umlcon_set_palette(struct vc_data *c, unsigned char *table);
static int umlcon_scrolldelta(struct vc_data *c, int lines);
static int umlcon_set_origin(struct vc_data *c);
static void umlcon_save_screen(struct vc_data *c);
static int umlcon_scroll(struct vc_data *c, int t, int b, int dir, int lines);
static void umlcon_bmove(struct vc_data *c, int sy, int sx, int dy, int dx,
		int height, int width);
static u8 umlcon_build_attr(struct vc_data *c, u8 color, u8 intensity, u8 blink, u8 underline, u8 reverse);
static void umlcon_invert_region(struct vc_data *c, u16 *p, int count);


static void umlcon_send(int fd, unsigned char *cmd)
{
	char errmsg[128];
	static int disable = 0;
	//int len;
	int ret;
	static char ypos = 0;

	if (disable)
		return;

	assert(FD_DISPLAY == fd);

	// UMLfront is not used anymore. So instead of sending packets which
	// would have been interpreted by it, we directly write characters to
	// stdout. We print newlines in between to make it readable -MAB

	do {
	        if (cmd[0] == 1) {
		        if (cmd[4] == 0 || cmd[3] != ypos) {
			        do {
				        ret = lotol_write(1, "\n", 1);
				} while (ret < 0 && (errno == LOTOL_EINTR ||
						     errno == LOTOL_EAGAIN));
				ypos = cmd[3];
			}
			ret = lotol_write(1, cmd+2, 1);
		}
		else
		        ret = 1;
	} while (ret < 0 && (errno == LOTOL_EINTR || errno == LOTOL_EAGAIN));
	if (ret < 0 && errno != LOTOL_EAGAIN) {
	        /* real problem */
	        sprintf(errmsg,
			"cannot send to console: errno=%d.\n", errno);
		lotol_write(1, errmsg, strlen(errmsg));
		//disable = 1;
		
	} else if (ret == 0) {
	        /* unknown problem */
	        sprintf(errmsg,
			"cannot send to console: short write.\n");
		lotol_write(1, errmsg, strlen(errmsg));
		//disable = 1;
	}
}

void no_scroll(char *str, int *ints)
{
}

static const char __init *umlcon_startup(void)
{
        //char errmsg[128];
        //int ret;

        if (ORIG_VIDEO_ISVGA == VIDEO_TYPE_VLFB) {
#ifdef CONFIG_DUMMY_CONSOLE
                conswitchp = &dummy_con;
                return conswitchp->con_startup();
#else
                return NULL;
#endif
	}
	/* Won't be using FD_DISPLAY anymore -MAB
        } else {
		// XXX ?
		ret = lotol_fcntl(FD_DISPLAY, LOTOL_F_GETOWN, 0);
		if (-1 == ret) {
			// Somethings wrong with FD_DISPLAY!
			sprintf(errmsg,
				"cannot do fcntl on FD_DISPLAY: errno=%d.\n",
				errno);
			lotol_write(1, errmsg, strlen(errmsg));
			// XXX Should we panic?
		}
	}
	*/
	return "UML";
}

static void umlcon_init(struct vc_data *c, int init)
{
	/* We cannot be loaded as a module, therefore init is always 1 */
	c->vc_can_do_color = 1;
	c->vc_cols = 80;
	c->vc_rows = 25;
	c->vc_complement_mask = 0x7700;
}

static void umlcon_deinit(struct vc_data *c)
{
}

static u8 umlcon_build_attr(struct vc_data *c, u8 color, u8 intensity, u8 blink, u8 underline, u8 reverse)
{
	u8 attr = color;

	if (underline)
		attr = (attr & 0xf0) | c->vc_ulcolor;
	else if (intensity == 0) {
		attr = (attr & 0xf0) | c->vc_halfcolor;
	}
	if (reverse)
		attr = ((attr) & 0x88) | ((((attr) >> 4) | ((attr) << 4)) & 0x77);
	if (blink)
		attr ^= 0x80;
	if (intensity == 2)
		attr ^= 0x08;

	return attr;
}

static void umlcon_invert_region(struct vc_data *c, u16 *p, int count)
{
	char str[100];

	sprintf(str, "%s: call(c=?, p=%p, count=%d)\n", __FUNCTION__, p, count);
	lotol_write(1, str, strlen(str));
}

static void umlcon_cursor(struct vc_data *c, int mode)
{
	unsigned char cmd[7];
	unsigned long flags;

	spin_lock_irqsave(&uml_lock, flags);

	cmd[0] = 4;	/* cursor */
	cmd[1] = mode;
	cmd[2] = c->vc_cursor_type & 0x0f;
	cmd[3] = c->vc_y;
	cmd[4] = c->vc_x;
	cmd[5] = 0;
	cmd[6] = 0;
	umlcon_send(FD_DISPLAY, cmd);

	spin_unlock_irqrestore(&uml_lock, flags);
}

static void umlcon_clear(struct vc_data *c, int sy, int sx, int height, int width)
{
	unsigned char cmd[7];
	unsigned long flags;

	spin_lock_irqsave(&uml_lock, flags);

	cmd[0] = 3;	/* clear */
	cmd[1] = sy;
	cmd[2] = sx;
	cmd[3] = height;
	cmd[4] = width;
	cmd[5] = (c->vc_video_erase_char >> 8) & 0xff;
	cmd[6] = (c->vc_video_erase_char >> 0) & 0xff;
	umlcon_send(FD_DISPLAY, cmd);

	spin_unlock_irqrestore(&uml_lock, flags);
}

static void umlcon_putc(struct vc_data *c, int ch, int ypos, int xpos)
{
	unsigned char cmd[7];
	unsigned long flags;

	spin_lock_irqsave(&uml_lock, flags);

	cmd[0] = 1;	/* putc */
	cmd[1] = (ch >> 8) & 0xff;
	cmd[2] = ch & 0xff;
	cmd[3] = ypos;
	cmd[4] = xpos;
	cmd[5] = 0;
	cmd[6] = 0;
	umlcon_send(FD_DISPLAY, cmd);

	spin_unlock_irqrestore(&uml_lock, flags);
}

static void umlcon_putcs(struct vc_data *c, const unsigned short *s, int count, int ypos, int xpos)
{
	for ( ; 0 < count; count--) {
		umlcon_putc(c, *s++, ypos, xpos++);
	}
}

static int umlcon_switch(struct vc_data *c)
{
	return 1;	/* Redrawing needed */
}

static int umlcon_set_palette(struct vc_data *c, unsigned char *table)
{
	return -EINVAL;
}

static int umlcon_blank(struct vc_data *c, int blank)
{
	return 0;
}

static int umlcon_font_op(struct vc_data *c, struct console_font_op *op)
{
	return -ENOSYS;
}

static int umlcon_scrolldelta(struct vc_data *c, int lines)
{
	char str[100];

	sprintf(str, "%s: call(c=?, lines=%d)\n", __FUNCTION__, lines);
	lotol_write(1, str, strlen(str));
	return 1;
}

static int umlcon_set_origin(struct vc_data *c)
{
	return 0;
}

static void umlcon_save_screen(struct vc_data *c)
{
}

static int umlcon_scroll(struct vc_data *c, int t, int b, int dir, int lines)
{
	unsigned char cmd[7];
	unsigned long flags;

	spin_lock_irqsave(&uml_lock, flags);


	cmd[0] = 2;	/* scroll */
	cmd[1] = t;
	cmd[2] = b;
	cmd[3] = dir;
	cmd[4] = lines;
	cmd[5] = (c->vc_video_erase_char >> 8) & 0xff;
	cmd[6] = (c->vc_video_erase_char >> 0) & 0xff;
	umlcon_send(FD_DISPLAY, cmd);

	spin_unlock_irqrestore(&uml_lock, flags);
	return 0;
}

static void umlcon_bmove(struct vc_data *conp, int sy, int sx, int dy, int dx, int height, int width)
{
	unsigned char cmd[7];
	unsigned long flags;

	spin_lock_irqsave(&uml_lock, flags);

	cmd[0] = 5;	/* bmove */
	cmd[1] = sy;
	cmd[2] = sx;
	cmd[3] = dy;
	cmd[4] = dx;
	cmd[5] = height;
	cmd[6] = width;
	umlcon_send(FD_DISPLAY, cmd);

	spin_unlock_irqrestore(&uml_lock, flags);
}

/*
 *  The console `switch' structure for the UMLinux based console
 */

const struct consw vga_con = {
	con_startup:		umlcon_startup,
	con_init:		umlcon_init,
	con_deinit:		umlcon_deinit,
	con_clear:		umlcon_clear,
	con_putc:		umlcon_putc,
	con_putcs:		umlcon_putcs,
	con_cursor:		umlcon_cursor,
	con_scroll:		umlcon_scroll,
	con_bmove:		umlcon_bmove,
	con_switch:		umlcon_switch,
	con_blank:		umlcon_blank,
	con_font_op:		umlcon_font_op,
	con_set_palette:	umlcon_set_palette,
	con_scrolldelta:	umlcon_scrolldelta,
	con_set_origin:		umlcon_set_origin,
	con_save_screen:	umlcon_save_screen,
	con_build_attr:		umlcon_build_attr,
	con_invert_region:	umlcon_invert_region
};
