/*
 * Network driver for linux system.
 * Copyright Volkmar Sieh / University Erlangen/Nueremberg
 */

static const char *version =
"UMN Version 1.0 by University Erlangen-Nuremberg\n";


#include <linux/init.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <linux/if_arp.h>	/* For ARPHRD_ETHER */

#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>

#define ETHER_ADDR_LEN	6


struct umn_private {
  //int fd;
  int tfd;
  int myself;
  struct {
    unsigned char mac[ETHER_ADDR_LEN];;
    long addr;
    int port;
  } host[16];
  struct net_device_stats stats;
};

static void umn_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  struct device *dev = (struct device *) dev_id;
  struct umn_private *umn = (struct umn_private *) dev->priv;
  
  static char buf[65536];
  int len;
  //struct lotol_sockaddr_in from;
  //int fromlen;
  struct sk_buff *skb;
  
  for (;;) {
    //fromlen = sizeof(from);
    //len = lotol_recvfrom(umn->fd, buf, sizeof(buf), 0,
    //		 (struct lotol_sockaddr *) &from, &fromlen);

    len = lotol_read(umn->tfd, buf, sizeof(buf));

    if (len < 0)
      break;

    //printk("%s: TAP: In umn_interrupt!.\n", dev->name);

    skb = dev_alloc_skb(len + 2);
    if (! skb) {
      printk(KERN_INFO "%s: Memory squeeze, deferring packet.\n",
	     dev->name);
      return;
    }
    
    skb->dev = dev;
    skb_reserve(skb, 2);		/* 16 byte align */
    skb_put(skb, len);			/* make room */
    eth_copy_and_sum(skb, buf, len, 0);
    skb->protocol = eth_type_trans(skb, dev);
    netif_rx(skb);
    
    umn->stats.rx_bytes += len;
    umn->stats.rx_packets++;
  }
}

static int umn_xmit(struct sk_buff *skb, struct device *dev)
{
  struct umn_private *umn = (struct umn_private *) dev->priv;
  struct ethhdr *hdr;
  struct lotol_sockaddr_in to;
  int i;
  int ret;

  assert(skb != NULL);
  assert(sizeof(*hdr) <= skb->len);
  
  dev->tbusy = 1;
  
  hdr = (struct ethhdr *) skb->data;
  /*
  for (i = 0; i < sizeof(umn->host) / sizeof(umn->host[0]); i++) {
    if (i == umn->myself) continue;
    if (umn->host[i].addr == 0
	&& umn->host[i].port == 0)
      continue;
    memset(&to, 0, sizeof(to));
    to.sin_family = LOTOL_AF_INET;
    to.sin_port = lotol_htons(umn->host[i].port);
    to.sin_addr = lotol_htonl(umn->host[i].addr);
    (void) lotol_sendto(umn->fd, skb->data, skb->len, 0,
    			(struct lotol_sockaddr *) &to,
    			sizeof(to));
  }
  */
  
  //printk("%s: TAP: In umn_xmit.\n", dev->name);

  ret = lotol_write(umn->tfd, skb->data, skb->len);

  if (ret < 0)
    printk("\terrno: %d\n", errno);
  /*
  printk("\t: TAP: Wrote %d bytes to %d\n", ret, umn->tfd);
  printk("\t: Data: (%d bytes) ", skb->len);
  for (i = 0; i < skb->len && i < 100; i++) {
    printk(" %02x", skb->data[i]);
  }
  printk("\n");
  */
  dev_kfree_skb(skb);
  
  dev->tbusy = 0;
  
  umn->stats.tx_bytes += skb->len;
  umn->stats.tx_packets++;
  
  return 0;
}

static struct net_device_stats *umn_get_stats(struct device *dev)
{
  struct umn_private *up = (struct umn_private *) dev->priv;
  
  return &up->stats;
}

static void umn_set_multicast(struct device *dev)
{
  /* FIXME */
}

static int umn_open(struct device *dev)
{
  if (request_irq(dev->irq, umn_interrupt, SA_SHIRQ,
		  dev->name, dev)) {
    printk(KERN_ERR "%s: Can't get irq %d\n",
	   dev->name, dev->irq);
    return -EAGAIN;
  }
  dev->tbusy = 0;
  dev->interrupt = 0;
  dev->start = 1;
  return 0;
}

static int umn_close(struct device *dev)
{
  dev->start = 0;
  dev->tbusy = 1;
  free_irq(dev->irq, dev);
  return 0;
}

__initfunc(int umn_probe(struct device *dev))
{
  static unsigned version_printed = 0;
  
  struct umn_private *umn;
  int i;
  int fd;
  char c;
  char offset;
  int ret;
  
  if (version_printed++ == 0)
    printk(version);
  
  /*
   * allocate private memory
   */
  umn = kmalloc(sizeof(struct umn_private), GFP_KERNEL);
  if (umn == NULL) {
    printk("%s: unable to get memory for dev->priv.\n", dev->name);
    return ENOMEM;
  }
  
  assert(strlen(dev->name) == 4);
  offset = dev->name[3] - '0';
  assert(offset <= 7);
  
  /*
   * read network description from file
   */
  fd = FD_ETH0_CONF + offset;
  ret = lotol_lseek(fd, 0, 0);
  if (ret < 0) {
    kfree(umn);
    return ENXIO;
  }
  ret = lotol_read(fd, &c, 1);
  
  if (ret < 0) {
    kfree(umn);
    return ENXIO;
  }
  /* read `myself' */
  umn->myself = 0;
  if (ret != 1) c = '\0';
  while ('0' <= c && c <= '9') {
    umn->myself *= 10;
    umn->myself += c - '0';
    ret = lotol_read(fd, &c, 1);
    if (ret != 1) c = '\0';
  }
  /* read mac address / ip address/port */
  for (i = 0; i < sizeof(umn->host) / sizeof(umn->host[0]); i++) {
    int j;
    
    /* read mac address */
    for (j = 0; j < ETHER_ADDR_LEN; j++) {
      /* read digits */
      umn->host[i].mac[j] = 0;
      ret = lotol_read(fd, &c, 1);
      if (ret != 1) c = '\0';
      while (('0' <= c && c <= '9')
	     || ('A' <= c && c <= 'F')
	     || ('a' <= c && c <= 'f')) {
	umn->host[i].mac[j] *= 16;
	if ('0' <= c && c <= '9')
	  umn->host[i].mac[j] += c - '0';
	else if ('A' <= c && c <= 'F')
	  umn->host[i].mac[j] += c - 'A' + 10;
	else if ('a' <= c && c <= 'f')
	  umn->host[i].mac[j] += c - 'a' + 10;
	ret = lotol_read(fd, &c, 1);
	if (ret != 1) c = '\0';
      }
    }
    
    /* read ip address */
    umn->host[i].addr = 0;
    ret = lotol_read(fd, &c, 1);
    if (ret != 1) c = '\0';
    while (('0' <= c && c <= '9')
	   || ('A' <= c && c <= 'F')
	   || ('a' <= c && c <= 'f')) {
      umn->host[i].addr *= 16;
      if ('0' <= c && c <= '9')
	umn->host[i].addr += c - '0';
      else if ('A' <= c && c <= 'F')
	umn->host[i].addr += c - 'A' + 10;
      else if ('a' <= c && c <= 'f')
	umn->host[i].addr += c - 'a' + 10;
      ret = lotol_read(fd, &c, 1);
      if (ret != 1) c = '\0';
    }
    
    /* read ip port */
    umn->host[i].port = 0;
    ret = lotol_read(fd, &c, 1);
    if (ret != 1) c = '\0';
    while (('0' <= c && c <= '9')
	   || ('A' <= c && c <= 'F')
	   || ('a' <= c && c <= 'f')) {
      umn->host[i].port *= 16;
      if ('0' <= c && c <= '9')
	umn->host[i].port += c - '0';
      else if ('A' <= c && c <= 'F')
	umn->host[i].port += c - 'A' + 10;
      else if ('a' <= c && c <= 'f')
	umn->host[i].port += c - 'a' + 10;
      ret = lotol_read(fd, &c, 1);
      if (ret != 1) c = '\0';
    }
  }
  
  /*
   * allocate resources
   */
  dev->irq = LOTOL_SIGIO + 16;
  dev->base_addr = (int) dev;	/* use some address */
  dev->dma = 0;
  
  /*
   * init device
   */

  /*
  umn->fd = FD_ETH0 + offset;
  
  ret = lotol_fcntl(umn->fd, LOTOL_F_SETOWN, lotol_getpid());
  if (ret < 0) {
    printk("%s: can't set owner\n", dev->name);
    // (void) lotol_close(umn->fd);
    return ENXIO;
  }
  */

  /*
   * TAP
   */
  
  umn->tfd = FD_TAP;

  ret = lotol_fcntl(umn->tfd, LOTOL_F_SETOWN, lotol_getpid());
  if (ret < 0) {
    printk("%s: TAP: can't set owner\n", dev->name);
    // (void) lotol_close(umn->tfd); 
    return ENXIO;
  }
    
  memset(&umn->stats, 0, sizeof(umn->stats));
  
  for (i = 0; i < ETHER_ADDR_LEN; i++) {
    dev->dev_addr[i] = umn->host[umn->myself].mac[i];
  }
  dev->priv = umn;
  dev->open = umn_open;
  dev->stop = umn_close;
  dev->hard_start_xmit = umn_xmit;
  dev->get_stats = umn_get_stats;
  dev->set_multicast_list = umn_set_multicast;
  
  ether_setup(dev);
  
  printk("%s: found (using IRQ %d, MAC", dev->name, dev->irq);
  for (i = 0; i < ETHER_ADDR_LEN; i++) {
    printk("%c%2.2x", i == 0 ? ' ' : ':',
	   umn->host[umn->myself].mac[i]);
  }
  printk(", port 0x%x)\n", umn->host[umn->myself].port);

  return 0;
}
