/*
 * analyze.c
 *
 * Tries to analyze the contents of recordsize and recorddata files lying in
 * the same directory.
 *
 * Prints a lot of output (intentionally), including destination and source,
 * MAC and IP addresses, per packet information and connection information.
 *
 * Can be used to find any errors in recorded packets. Error/Warnings is sent
 * to stderr while other info is sent to stdout.
 *
 */

/* Hack to get ntohl etc working */
#define _LINUX_BYTEORDER_GENERIC_H	1

#define _GNU_SOURCE

#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <fcntl.h>


volatile int term = 0;

FILE *fp1, *fp2;

char *sizes;

unsigned lost_bytes, repeated_bytes;

struct {
  unsigned char port[2];
  unsigned seq;
} connection[64];


void
int_handler(int sig)
{
  fprintf(stderr, "SIGNAL: going down on signal %d : %s\n", sig,
	  strsignal(sig));
  term = 1;
}


void
dump(char *string, unsigned char *buf, int bufsize)
{
  int i;

  fprintf(stdout, "%s: \n", string);
  for (i = 0; i < bufsize && i < 100; i++) {
    fprintf(stdout, "%02x ", buf[i]);
  }
  fprintf(stdout, "\n");
}


void
openfile(void)
{
  unsigned size;
  if( (fp1 = fopen("./recordsize", "r")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./recordsize\n");
    term = 1;
  }
  
  fseek(fp1, 0, SEEK_END);
  size = ftell(fp1);
  fseek(fp1, 0, SEEK_SET); 
  sizes = (char *)malloc(size);
  if( (fread((void *)sizes, 1, size, fp1)) != size ){
    fprintf(stderr, "FILEIO: Error in reading from file ./recordsize\n");
    term = 1;
  }
  fclose(fp1);
  sizes[size-1] = '\0';

  if( (fp2 = fopen("./recorddata", "r")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./recorddata\n");
    term = 1;
  }
}


void
closefile(void)
{
  fclose(fp2);
}


unsigned char *
readrecord(int recordsize)
{
  unsigned char *buf = (unsigned char *)malloc(recordsize);
  if( (fread((void *)buf, 1, recordsize, fp2)) 
      != recordsize ){
    fprintf(stderr, "FILEIO: Error in reding from file ./recorddata\n");
    term = 1;
  }
  return buf;
}


/*
 * port2sock manages the ports in the connection array.
 * action = 0 : check whether the port exists.
 *          1 : add the port to the connection array.
 *          2 : delete the port from the connection array.
 */
int
port2sock(unsigned char *src_port, int action)
{
  int i;
  static unsigned char zero[2] = { 0, 0 };


  switch (action) {

  case 0:
    for( i = 0; i < 64; i++) {
      if( memcmp(src_port, connection[i].port , 2) == 0 ) {
	fprintf(stdout, "SOCK: port2sock - check\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
	return i;
      }
    }
    fprintf(stdout, "SOCK: port2sock - check\n");
    fprintf(stdout, "\t src_port: %02x %02x Check Failed.\n", src_port[0],
	    src_port[1]);
    return -1;

  case 1:
    for( i = 0; i < 64; i++) {
      if( memcmp(connection[i].port, zero, 2) == 0 ) {
	memcpy(connection[i].port, src_port, 2);
	fprintf(stdout, "SOCK: port2sock - add\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
	return i;
      }
    }
    fprintf(stdout, "SOCK: port2sock - add\n");
    fprintf(stdout, "\t src_port: %02x %02x Add Failed.\n", src_port[0],
	    src_port[1]);
    return -1;    

  case 2:
    for( i = 0; i < 64; i++) {
      if( memcmp(src_port, connection[i].port, 2) == 0 ) {
	memcpy(connection[i].port, zero, 2);
	fprintf(stdout, "SOCK: port2sock - del\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
	return i;
      }
    }
    fprintf(stdout, "SOCK: port2sock - del\n");
    fprintf(stdout, "\t src_port: %02x %02x Del Failed.\n", src_port[0],
	    src_port[1]);
    return -1;
  }
  
  return -1;
}


/*
 * seqmismatch prints warning messages and updates counters when sequence
 * numbers in packets are not as expected
 */
void
seqmismatch(unsigned expected, unsigned obtained)
{
  fprintf(stderr, "PARSE: WARNING: Sequence number mismatch. Expected"
	  " %u got %u.\n", expected, obtained);
  fprintf(stdout, "PARSE: WARNING: Sequence number mismatch. Expected"
	  " %u got %u.\n", expected, obtained);
  if( expected < obtained ) {
    fprintf(stderr, "\tMissed %u bytes.\n", obtained - expected);
    fprintf(stdout, "\tMissed %u bytes.\n", obtained - expected);
    lost_bytes += obtained - expected;
  }
  else {
    fprintf(stderr, "\tRepeated %u bytes.\n", expected - obtained);
    fprintf(stdout, "\tRepeated %u bytes.\n", expected - obtained);
    repeated_bytes += expected - obtained;
  }
}


/*
 * parserecord understand the contents of the packet and does the needfull.
 * 
 * The frame_type and destination_port should always be as expected as the 
 * recording function takes care of this.
 * Assumptions are made about ihl, version, fragmentation and protocol fields
 * and a debugging message will be sent to stderr if they are broken.
 * Destination and source mac and ip addresses are expected to be the same, and
 * any changes will be logged.
 *
 * Optional TCP headers are not explicitely taken care of, but the correct TCP
 * data is found and sent using write.
 * Socket management is done when a syn, fin or res packet is encountered.
 * Otherwise if the packet contains TCP data, it is written to the appropriate
 * connection (if it exists). The pipe is always emptied by reading data from 
 * it before writing anything to it.
 *
 * NOTE: This does not start any TCP connections or write to them!
 * It just tries to simulate what would happen without doing anything.
 */
void
parserecord(unsigned char *buf, int recordsize)
{
  unsigned char control;
  int urg, ack, pus, res, syn, fin;
  unsigned char tcp_header_len;
  int tcp_header_length;
  int data_size;
  unsigned seq_no;
  int sockno;
  static int i = 0;


  // For recording MAC address of Destination[6].
  static unsigned char dst_mac[6] = {
    0, 0, 0, 0, 0, 0
  };
  // For recording MAC address of Source[6].
  static unsigned char src_mac[6] = {
    0, 0, 0, 0, 0, 0
  };
  // Frame type[2] for IP frames.
  static unsigned char frame_type_ip[2] = {
    0x08, 0
  };
  // Version + Header Length[1], Type of Service[1].
  static unsigned char misc1[2] = {
    0x45, 0
  };
  // Skip Length[2], Id[2].
  // Fragmentation offset[2] for no fragmentation.
  static unsigned char frag[2] = {
    0x40, 0
  };
  // Skip Time to live[1].
  // Protocol[1] for TCP.
  static unsigned char proto[1] = {
    0x06
  };
  // Skip Checksum[2].
  // For recording IP address of Source[4].
  static unsigned char src_ip[4] = {
    0, 0, 0, 0
  };
  // For recording IP address of Destination[4].
  static unsigned char dst_ip[4] = {
    0, 0, 0, 0
  };
  // Skip Source port[2].
  // Destination port[2]. (6000)
  static unsigned char dst_port[2] = {
    0x17, 0x70
  };
  // Skip Sequence no[4], Acknowledgement no[4].

  if ( memcmp(buf + 12, frame_type_ip, 2) != 0
       || memcmp(buf + 36, dst_port, 2) != 0
       ) {
    // Panic.
    fprintf(stderr, "PARSE: Error in parsing headers\n");
    dump("error", buf, recordsize);
    term = 1;
    return;
  }

  if ( memcmp(buf + 14, misc1, 2) != 0
       || memcmp(buf + 20, frag, 2) != 0
       || memcmp(buf + 23, proto, 1) != 0
       ) {
    // Concern.
    fprintf(stderr, "PARSE: WARNING: Trouble in parsing headers\n");
    dump("warning", buf, recordsize);
    return;
  }

  if ( memcmp(buf, dst_mac, 6) != 0 ) {
    // New Destination MAC address.
    fprintf(stdout, "PARSE: DST_MAC: %02x %02x %02x %02x %02x %02x\n",
	    buf[0], buf[1], buf[2], buf[3], buf[4], buf[5]);
    memcpy(dst_mac, buf, 6);
  }
  if ( memcmp(buf + 6, src_mac, 6) != 0 ) {
    // New Source MAC address.
    fprintf(stdout, "PARSE: SRC_MAC: %02x %02x %02x %02x %02x %02x\n",
	    buf[6+0], buf[6+1], buf[6+2], buf[6+3], buf[6+4], buf[6+5]);
    memcpy(src_mac, buf + 6, 6);
  }
  if ( memcmp(buf + 26, src_ip, 4) != 0 ) {
    // New Source IP address.
    fprintf(stdout, "PARSE: SRC_IP: %02x %02x %02x %02x\n",
	    buf[26+0], buf[26+1], buf[26+2], buf[26+3]);
    memcpy(src_ip, buf + 26, 4);
  }
  if ( memcmp(buf + 30, dst_ip, 4) != 0 ) {
    // New Destination IP address.
    fprintf(stdout, "PARSE: DST_IP: %02x %02x %02x %02x\n",
	    buf[30+0], buf[30+1], buf[30+2], buf[30+3]);
    memcpy(dst_ip, buf + 30, 4);
  }


  // Extract Offset[4/8]
  tcp_header_len = (buf[46] & 0xf0) >> 4;
  tcp_header_length = tcp_header_len*4;
  data_size = recordsize-34-tcp_header_length;

  // Skip Reserved[6/8]
  // Extract Control Flags[6/8]
  control = buf[47];
  urg = (control & 0x20) >> 5;
  ack = (control & 0x10) >> 4;
  pus = (control & 0x08) >> 3;
  res = (control & 0x04) >> 2;
  syn = (control & 0x02) >> 1;
  fin = control & 0x01;

  seq_no = buf[38+0]*256u*256u*256u + buf[38+1]*256u*256u + buf[38+2]*256u 
           + buf[38+3];

  fprintf(stdout, "\nPARSE: Info about packet %d\n", i);

  fprintf(stdout, "   IP LAYER: Length: %02x %02x %4u \t       Id: %02x %02x"
	  " %5u \n",
	  buf[16+0], buf[16+1], buf[16+0]*256u + buf[16+1],
	  buf[18+0], buf[18+1], buf[18+0]*256u + buf[18+1]);
  fprintf(stdout, "  TCP LAYER: Seq No: %02x %02x %02x %02x %10u  "
	  "  Ack No: %02x %02x %02x %02x %10u \n",
	  buf[38+0], buf[38+1], buf[38+2], buf[38+3], seq_no,
	  buf[42+0], buf[42+1], buf[42+2], buf[42+3], buf[42+0]*256u*256u*256u
	  + buf[42+1]*256u*256u + buf[42+2]*256u + buf[42+3]);
  fprintf(stdout, "\tTCP Header Length: %d    FLAGS: Urg: %d Ack: %d Pus: %d "
	  "Res %d Syn %d Fin %d\n", tcp_header_length, urg, ack, pus, res, syn,
	  fin);


  if(syn) {

    fprintf(stdout, "PARSE: Starting syn\n");
    dump("syn", buf, recordsize);

    // Close socket if connection exists.
    sockno = port2sock(buf+34, 0);
    if ( sockno != -1 ) {
      port2sock(buf+34, 2);
    }

    // Prepare for new connection.
    sockno = port2sock(buf+34, 1);
    if ( sockno == -1 ) {
      fprintf(stderr, "PARSE: Exceeded max sockets\n");
      term = 1;
      return;
    }

    // Set sequence number to that of next packet to be sent on this
    // connection.
    connection[sockno].seq = seq_no + 1;

  }
  else if(fin) {

    fprintf(stdout, "PARSE: Starting fin\n");
    dump("fin", buf, recordsize);

    // Close socket if connection exists.
    sockno = port2sock(buf+34, 0);
    if ( sockno != -1 ) {
      // Check if sequence number is as expected.
      if( connection[sockno].seq != seq_no ) {
	seqmismatch(connection[sockno].seq, seq_no);
      }
      port2sock(buf+34, 2);
      connection[sockno].seq = 0; 
    }
    else {
      fprintf(stdout, "PARSE: No connection to close.\n");
    }

  }
  else if(res) {

    fprintf(stdout, "PARSE: Connection reset by Peer\n");
    dump("reset", buf, recordsize);

    // Close socket if connection exists.
    sockno = port2sock(buf+34, 0);
    if ( sockno != -1 ) {
      // Check if sequence number is as expected.
      if( connection[sockno].seq != seq_no ) {
	seqmismatch(connection[sockno].seq, seq_no);
      }
      port2sock(buf+34, 2);
      connection[sockno].seq = 0; 
    }
    else {
      fprintf(stdout, "PARSE: No connection to reset.\n");
    }

  }
  else {
	  
    if( data_size > 0 ) {
      sockno = port2sock(buf+34, 0);
      if ( sockno != -1 ) {
	// Check if sequence number is as expected.
	if( connection[sockno].seq != seq_no ) {
	  seqmismatch(connection[sockno].seq, seq_no);
	}
	fprintf(stdout, "PARSE: Before writing packet %d.\n", i);
	dump("write", buf, recordsize);
	// Set sequence number to that of next packet to be sent on this
	// connection.
	connection[sockno].seq = seq_no + data_size; 
      }
      else {
	fprintf(stdout, "PARSE: No connection. Not writing packet.\n");
	dump("write", buf, recordsize);
      }
    }
    else {
      fprintf(stdout, "PARSE: Packet without data portion.\n");
      dump("ignore", buf, recordsize);
    }
    
  }
  fprintf(stdout, "PARSE: Record size: %4d   Data size: %4d\n", recordsize,
	  data_size);
  i++;
}


int
main(int argc, char *argv[] )
{
  struct sigaction sa;
  int ret;
  char *token;
  int recordsize;
  unsigned char *buffer;


  sa.sa_handler = int_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  ret = sigaction(SIGINT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGQUIT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGHUP, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGPIPE, &sa, (struct sigaction *) 0);
  assert(0 <= ret);


  openfile();

  if( ( token = strtok(sizes, " ") ) != NULL ) {
    recordsize = atoi(token);
    buffer = readrecord(recordsize);
    parserecord(buffer, recordsize);
    free(buffer);
  }

  while(!term) {
    if( ( token = strtok(NULL, " ") ) != NULL ) {
      recordsize = atoi(token);
      buffer = readrecord(recordsize);
      parserecord(buffer, recordsize);
      free(buffer);
    }
    else
      term = 1;
  }

  closefile();

  if( lost_bytes > 0 ) {
    fprintf(stderr, "Lost %u bytes totally.\n", lost_bytes);
    fprintf(stdout, "Lost %u bytes totally.\n", lost_bytes);
  }
  if( repeated_bytes > 0 ) {
    fprintf(stderr, "Repeated %u bytes totally.\n", repeated_bytes);
    fprintf(stdout, "Repeated %u bytes totally.\n", repeated_bytes);
  }

  return 0;
}
