/*
 * ipt_user_record.c
 *
 * Acts as a user-space application for IPTABLES and gives an ACCEPT verdict
 * for all the packets passed to it.
 * It also records the sizes of the packets in iprecordsize and the packet data
 * in iprecorddata in the current directory.
 *
 */

/* Hack to get ntohl etc working */
#define _LINUX_BYTEORDER_GENERIC_H	1

#define _GNU_SOURCE

#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/netfilter.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include "libipq.h"

#define BUFSIZE 65536


volatile int term = 0;

FILE *fp1, *fp2;

struct {
  int sock;
  unsigned char port[2];
  unsigned seq;
} connection[64];


void
int_handler(int sig)
{
  switch(sig) {
  case SIGPIPE:
	fprintf(stderr, "SIGNAL: Caught SIGPIPE, ignoring.\n");
	break;
  default:
    fprintf(stderr, "SIGNAL: going down on signal %d : %s\n", sig,
	    strsignal(sig));
    term = 1;
  }
}


void
openfile(void)
{
  if( (fp1 = fopen("./iprecordsize", "w+")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./iprecordsize");
    term = 1;
  }
  if( (fp2 = fopen("./iprecorddata", "w+")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./iprecorddata");
    term = 1;
  }
}


void
closefile(void)
{
  fclose(fp1);
  fclose(fp2);
}


void
record(unsigned char *buf, int bufsize)
{
  char buffersize[16];
  sprintf(buffersize, "%d ", bufsize);
  if( (fwrite((void *)buffersize, 1, strlen(buffersize), fp1)) 
      != strlen(buffersize) ){
    fprintf(stderr, "FILEIO: Error in writing to file ./iprecordsize");
    fclose(fp1);
    term = 1;
  }

  if( (fwrite((void *)buf, 1, bufsize, fp2)) 
      != bufsize ){
    fprintf(stderr, "FILEIO: Error in writing to file ./iprecorddata");
    fclose(fp2);
    term = 1;
  }
}


void
dump(char *string, unsigned char *buf, int bufsize)
{
  int i;

  fprintf(stdout, "%s: \n", string);
  for (i = 0; i < bufsize && i < 100; i++) {
    fprintf(stdout, "%02x ", buf[i]);
  }
  fprintf(stdout, "\n");
}


static void die(struct ipq_handle *h)
{
  ipq_perror("passer");
  ipq_destroy_handle(h);
  exit(1);
}


int main(int argc, char **argv)
{
  struct sigaction sa;
  int ret;
  int status;
  unsigned char buf[BUFSIZE];
  struct ipq_handle *h;


  sa.sa_handler = int_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  ret = sigaction(SIGINT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGQUIT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGHUP, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGPIPE, &sa, (struct sigaction *) 0);
  assert(0 <= ret);

  openfile();

  h = ipq_create_handle(0);
  if (!h)
    die(h);
  
  status = ipq_set_mode(h, IPQ_COPY_PACKET, BUFSIZE);
  if (status < 0)
    die(h);
  
  do{
    status = ipq_read(h, buf, BUFSIZE, 0);
    if (status < 0)
      die(h);
    
    switch (ipq_message_type(buf)) {
    case NLMSG_ERROR:
      fprintf(stderr, "Received error message %d\n", ipq_get_msgerr(buf));
      break;
      
    case IPQM_PACKET: {
      ipq_packet_msg_t *m = ipq_get_packet(buf);

      record(m->payload, m->data_len);
            
      status = ipq_set_verdict(h, m->packet_id, NF_ACCEPT, 0, NULL);
      if (status < 0)
	die(h);

      break;
    }
    
    default:
      fprintf(stderr, "Unknown message type!\n");
      break;
    }
  } while (!term);
  
  closefile();

  ipq_destroy_handle(h);
  return 0;
}
